"""
These widget objects provide the interaction to yt tasks.

Author: Samuel Skillman <samskillman@gmail.com>
Affiliation: University of Colorado at Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from yt.mods import *
import weakref

class RenderingScene(object):
    _camera = None
    _tf = None
    
    def __init__(self, pf, camera=None, tf=None):
        self.pf = weakref.proxy(pf)
        self._camera = camera
        self._tf = tf

        self.center = self.pf.domain_center
        self.normal_vector = np.array([0.7,1.0,0.3])
        self.north_vector = [0.,0.,1.]
        self.steady_north = True
        self.fields = ['Density']
        self.log_fields = [True]
        self.l_max = 0
        self.mi = None
        self.ma = None
        if self._tf is None:
            self._new_tf(self.pf)

        if self._camera is None:
            self._new_camera(self.pf)

    def _new_tf(self, pf, mi=None, ma=None, nbins=1024):
        if mi is None or ma is None:
            roi = self.pf.h.region(self.center, self.center-self.width, self.center+self.width)
            self.mi, self.ma = roi.quantities['Extrema'](self.fields[0])[0]
            if self.log_fields[0]:
                self.mi, self.ma = np.log10(self.mi), np.log10(self.ma)

        self._tf = ColorTransferFunction((self.mi-2, self.ma+2), nbins=nbins)

    def add_contours(self, n_contours=7, contour_width=0.05, colormap='kamae'):
        self._tf.add_layers(n_contours=n_contours ,w=contour_width,
                                  col_bounds = (self.mi,self.ma), 
                                  colormap=colormap)

    def _new_camera(self, pf):
        del self._camera
        self._camera = self.pf.camera(self.center, self.normal_vector, 
                                      self.width, self.resolution, self._tf,
                                      north_vector=self.north_vector,
                                      steady_north=self.steady_north,
                                      fields=self.fields, log_fields=self.log_fields,
                                      l_max=self.l_max)
    def snapshot(self):
        return self._camera.snapshot()


def get_corners(pf, max_level=None):
    DL = pf.domain_left_edge[None,:,None]
    DW = pf.domain_width[None,:,None]/100.0
    corners = ((pf.h.grid_corners-DL)/DW)
    levels = pf.h.grid_levels
    return corners, levels

def get_isocontour(pf, field, value=None, rel_val = False):

    dd = pf.h.all_data()
    if value is None or rel_val:
        if value is None: value = 0.5
        mi, ma = np.log10(dd.quantities["Extrema"]("Density")[0])
        value = 10.0**(value*(ma - mi) + mi)
    vert = dd.extract_isocontours("Density", value)
    np.multiply(vert, 100, vert)
    return vert

def get_streamlines(pf):
    from yt.visualization.api import Streamlines
    streamlines = Streamlines(pf, pf.domain_center) 
    streamlines.integrate_through_volume()
    stream = streamlines.path(0)
    matplotlib.pylab.semilogy(stream['t'], stream['Density'], '-x')





"""
The data-file handling functions

Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import h5py
import os
import re
import numpy as np

from yt.utilities.io_handler import \
           BaseIOHandler

class IOHandlerChomboHDF5(BaseIOHandler):
    _data_style = "chombo_hdf5"
    _offset_string = 'data:offsets=0'
    _data_string = 'data:datatype=0'

    def __init__(self, pf, *args, **kwargs):
        BaseIOHandler.__init__(self, *args, **kwargs)
        self.pf = pf
        self._handle = pf._handle

    _field_dict = None
    @property
    def field_dict(self):
        if self._field_dict is not None:
            return self._field_dict
        ncomp = int(self._handle['/'].attrs['num_components'])
        temp =  self._handle['/'].attrs.items()[-ncomp:]
        val, keys = zip(*temp)
        val = [int(re.match('component_(\d+)',v).groups()[0]) for v in val]
        self._field_dict = dict(zip(keys,val))
        return self._field_dict
        
    def _read_field_names(self,grid):
        ncomp = int(self._handle['/'].attrs['num_components'])

        fns = [c[1] for c in f['/'].attrs.items()[-ncomp-1:-1]]
    
    def _read_data(self,grid,field):

        lstring = 'level_%i' % grid.Level
        lev = self._handle[lstring]
        dims = grid.ActiveDimensions
        boxsize = dims.prod()
        
        grid_offset = lev[self._offset_string][grid._level_id]
        start = grid_offset+self.field_dict[field]*boxsize
        stop = start + boxsize
        data = lev[self._data_string][start:stop]
        
        return data.reshape(dims, order='F')

    def _read_particles(self, grid, field):
        """
        parses the Orion Star Particle text files
             
        """

        fn = grid.pf.fullplotdir[:-4] + "sink"

        # Figure out the format of the particle file
        with open(fn, 'r') as f:
            lines = f.readlines()
        line = lines[1]

        # The basic fields that all sink particles have
        index = {'particle_mass': 0,
                 'particle_position_x': 1,
                 'particle_position_y': 2,
                 'particle_position_z': 3,
                 'particle_momentum_x': 4,
                 'particle_momentum_y': 5,
                 'particle_momentum_z': 6,
                 'particle_angmomen_x': 7,
                 'particle_angmomen_y': 8,
                 'particle_angmomen_z': 9,
                 'particle_id': -1}

        if len(line.strip().split()) == 11:
            # these are vanilla sinks, do nothing
            pass  

        elif len(line.strip().split()) == 17:
            # these are old-style stars, add stellar model parameters
            index['particle_mlast']     = 10
            index['particle_r']         = 11
            index['particle_mdeut']     = 12
            index['particle_n']         = 13
            index['particle_mdot']      = 14,
            index['particle_burnstate'] = 15

        elif len(line.strip().split()) == 18:
            # these are the newer style, add luminosity as well
            index['particle_mlast']     = 10
            index['particle_r']         = 11
            index['particle_mdeut']     = 12
            index['particle_n']         = 13
            index['particle_mdot']      = 14,
            index['particle_burnstate'] = 15,
            index['particle_luminosity']= 16

        else:
            # give a warning if none of the above apply:
            mylog.warning('Warning - could not figure out particle output file')
            mylog.warning('These results could be nonsense!')
            
        def read(line, field):
            return float(line.strip().split(' ')[index[field]])

        fn = grid.pf.fullplotdir[:-4] + "sink"
        with open(fn, 'r') as f:
            lines = f.readlines()
            particles = []
            for line in lines[1:]:
                if grid.NumberOfParticles > 0:
                    coord = read(line, "particle_position_x"), \
                            read(line, "particle_position_y"), \
                            read(line, "particle_position_z")
                    if ( (grid.LeftEdge < coord).all() and
                         (coord <= grid.RightEdge).all() ):
                        particles.append(read(line, field))
        return np.array(particles)



"""
FLASH-specific IO functions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import h5py

from yt.utilities.io_handler import \
    BaseIOHandler

class IOHandlerFLASH(BaseIOHandler):
    _particle_reader = False
    _data_style = "flash_hdf5"

    def __init__(self, pf, *args, **kwargs):
        self._num_per_stride = kwargs.pop("num_per_stride", 1000000)
        BaseIOHandler.__init__(self, *args, **kwargs)
        # Now we cache the particle fields
        self.pf = pf
        self._handle = pf._handle
        self._particle_handle = pf._particle_handle
        
        try :
            particle_fields = [s[0].strip() for s in
                               self._particle_handle["/particle names"][:]]
            self._particle_fields = dict([("particle_" + s, i) for i, s in
                                          enumerate(particle_fields)])
        except KeyError:
            self._particle_fields = {}

    def _read_particles(self, fields_to_read, type, args, grid_list,
            count_list, conv_factors):
        pass

    def _read_data(self, grid, field):
        f = self._handle
        f_part = self._particle_handle
        if field in self._particle_fields:
            if grid.NumberOfParticles == 0: return np.array([], dtype='float64')
            start = self.pf.h._particle_indices[grid.id - grid._id_offset]
            end = self.pf.h._particle_indices[grid.id - grid._id_offset + 1]
            fi = self._particle_fields[field]
            tr = f_part["/tracer particles"][start:end, fi]
        else:
            tr = f["/%s" % field][grid.id - grid._id_offset,:,:,:].transpose()
        return tr.astype("float64")



"""
Data structures for Athena.

Author: Samuel W. Skillman <samskillman@gmail.com>
Affiliation: University of Colorado at Boulder
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: John ZuHone <jzuhone@gmail.com>
Affiliation: NASA/Goddard Space Flight Center
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2013 Samuel W. Skillman, Matthew Turk, J. S. Oishi.,
  John ZuHone.
  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py
import numpy as np
import weakref
import glob #ST 9/12
from yt.funcs import *
from yt.data_objects.grid_patch import \
           AMRGridPatch
from yt.data_objects.hierarchy import \
           AMRHierarchy
from yt.data_objects.static_output import \
           StaticOutput
from yt.utilities.definitions import \
    mpc_conversion, sec_conversion

from .fields import AthenaFieldInfo, KnownAthenaFields
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc

def _get_convert(fname):
    def _conv(data):
        return data.convert(fname)
    return _conv

class AthenaGrid(AMRGridPatch):
    _id_offset = 0
    def __init__(self, id, hierarchy, level, start, dimensions):
        df = hierarchy.parameter_file.filename[4:-4]
        gname = hierarchy.grid_filenames[id]
        AMRGridPatch.__init__(self, id, filename = gname,
                              hierarchy = hierarchy)
        self.filename = gname
        self.Parent = []
        self.Children = []
        self.Level = level
        self.start_index = start.copy()
        self.stop_index = self.start_index + dimensions
        self.ActiveDimensions = dimensions.copy()

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        if len(self.Parent) > 0:
            self.dds = self.Parent[0].dds / self.pf.refine_by
        else:
            LE, RE = self.hierarchy.grid_left_edge[id,:], \
                     self.hierarchy.grid_right_edge[id,:]
            self.dds = np.array((RE-LE)/self.ActiveDimensions)
        if self.pf.dimensionality < 2: self.dds[1] = 1.0
        if self.pf.dimensionality < 3: self.dds[2] = 1.0
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

    def __repr__(self):
        return "AthenaGrid_%04i (%s)" % (self.id, self.ActiveDimensions)

def parse_line(line, grid):
    # grid is a dictionary
    splitup = line.strip().split()
    if "vtk" in splitup:
        grid['vtk_version'] = splitup[-1]
    elif "Really" in splitup:
        grid['time'] = splitup[-1]
    elif any(x in ['PRIMITIVE','CONSERVED'] for x in splitup):
        grid['time'] = float(splitup[4].rstrip(','))
        grid['level'] = int(splitup[6].rstrip(','))
        grid['domain'] = int(splitup[8].rstrip(','))
    elif "DIMENSIONS" in splitup:
        grid['dimensions'] = np.array(splitup[-3:]).astype('int')
    elif "ORIGIN" in splitup:
        grid['left_edge'] = np.array(splitup[-3:]).astype('float64')
    elif "SPACING" in splitup:
        grid['dds'] = np.array(splitup[-3:]).astype('float64')
    elif "CELL_DATA" in splitup:
        grid["ncells"] = int(splitup[-1])
    elif "SCALARS" in splitup:
        field = splitup[1]
        grid['read_field'] = field
        grid['read_type'] = 'scalar'
    elif "VECTORS" in splitup:
        field = splitup[1]
        grid['read_field'] = field
        grid['read_type'] = 'vector'

class AthenaHierarchy(AMRHierarchy):

    grid = AthenaGrid
    _data_style='athena'
    _data_file = None
    
    def __init__(self, pf, data_style='athena'):
        self.parameter_file = weakref.proxy(pf)
        self.directory = os.path.dirname(self.parameter_file.filename)
        self.data_style = data_style
        # for now, the hierarchy file is the parameter file!
        self.hierarchy_filename = self.parameter_file.filename
        #self.directory = os.path.dirname(self.hierarchy_filename)
        self._fhandle = file(self.hierarchy_filename,'rb')
        AMRHierarchy.__init__(self, pf, data_style)

        self._fhandle.close()

    def _detect_fields(self):
        field_map = {}
        f = open(self.hierarchy_filename,'rb')
        line = f.readline()
        while line != '':
            splitup = line.strip().split()
            if "DIMENSIONS" in splitup:
                grid_dims = np.array(splitup[-3:]).astype('int')
                line = f.readline()
            elif "CELL_DATA" in splitup:
                grid_ncells = int(splitup[-1])
                line = f.readline()
                if np.prod(grid_dims) != grid_ncells:
                    grid_dims -= 1
                    grid_dims[grid_dims==0]=1
                if np.prod(grid_dims) != grid_ncells:
                    mylog.error('product of dimensions %i not equal to number of cells %i' %
                          (np.prod(grid_dims), grid_ncells))
                    raise TypeError
                break
            else:
                line = f.readline()
        read_table = False
        read_table_offset = f.tell()
        while line != '':
            splitup = line.strip().split()
            if 'SCALARS' in splitup:
                field = splitup[1]
                if not read_table:
                    line = f.readline() # Read the lookup table line
                    read_table = True
                field_map[field] = ('scalar', f.tell() - read_table_offset)
                read_table=False

            elif 'VECTORS' in splitup:
                field = splitup[1]
                for ax in 'xyz':
                    field_map["%s_%s" % (field, ax)] =\
                            ('vector', f.tell() - read_table_offset)
            line = f.readline()

        f.close()

        self.field_list = field_map.keys()
        self._field_map = field_map

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

    def _count_grids(self):
        self.num_grids = self.parameter_file.nvtk

    def _parse_hierarchy(self):
        f = open(self.hierarchy_filename,'rb')
        grid = {}
        grid['read_field'] = None
        grid['read_type'] = None
        table_read=False
        line = f.readline()
        while grid['read_field'] is None:
            parse_line(line, grid)
            if "SCALAR" in line.strip().split():
                break
            if "VECTOR" in line.strip().split():
                break
            if 'TABLE' in line.strip().split():
                break
            if len(line) == 0: break
            line = f.readline()
        f.close()

        # It seems some datasets have a mismatch between ncells and 
        # the actual grid dimensions.
        if np.prod(grid['dimensions']) != grid['ncells']:
            grid['dimensions'] -= 1
            grid['dimensions'][grid['dimensions']==0]=1
        if np.prod(grid['dimensions']) != grid['ncells']:
            mylog.error('product of dimensions %i not equal to number of cells %i' % 
                  (np.prod(grid['dimensions']), grid['ncells']))
            raise TypeError

        # Need to determine how many grids: self.num_grids
        dname = self.hierarchy_filename
        gridlistread = glob.glob('id*/%s-id*%s' % (dname[4:-9],dname[-9:] ))
        gridlistread.insert(0,self.hierarchy_filename)
        if 'id0' in dname :
            gridlistread += glob.glob('id*/lev*/%s*-lev*%s' % (dname[4:-9],dname[-9:]))
        else :
            gridlistread += glob.glob('lev*/%s*-lev*%s' % (dname[:-9],dname[-9:]))
        self.num_grids = len(gridlistread)
        dxs=[]
        self.grids = np.empty(self.num_grids, dtype='object')
        levels = np.zeros(self.num_grids, dtype='int32')
        glis = np.empty((self.num_grids,3), dtype='float64')
        gdds = np.empty((self.num_grids,3), dtype='float64')
        gdims = np.ones_like(glis)
        j = 0
        self.grid_filenames = gridlistread
        while j < (self.num_grids):
            f = open(gridlistread[j],'rb')
            gridread = {}
            gridread['read_field'] = None
            gridread['read_type'] = None
            table_read=False
            line = f.readline()
            while gridread['read_field'] is None:
                parse_line(line, gridread)
                if "SCALAR" in line.strip().split():
                    break
                if "VECTOR" in line.strip().split():
                    break 
                if 'TABLE' in line.strip().split():
                    break
                if len(line) == 0: break
                line = f.readline()
            f.close()
            levels[j] = gridread['level']
            glis[j,0] = gridread['left_edge'][0]
            glis[j,1] = gridread['left_edge'][1]
            glis[j,2] = gridread['left_edge'][2]
            # It seems some datasets have a mismatch between ncells and 
            # the actual grid dimensions.
            if np.prod(gridread['dimensions']) != gridread['ncells']:
                gridread['dimensions'] -= 1
                gridread['dimensions'][gridread['dimensions']==0]=1
            if np.prod(gridread['dimensions']) != gridread['ncells']:
                mylog.error('product of dimensions %i not equal to number of cells %i' % 
                      (np.prod(gridread['dimensions']), gridread['ncells']))
                raise TypeError
            gdims[j,0] = gridread['dimensions'][0]
            gdims[j,1] = gridread['dimensions'][1]
            gdims[j,2] = gridread['dimensions'][2]
            # Setting dds=1 for non-active dimensions in 1D/2D datasets
            gridread['dds'][gridread['dimensions']==1] = 1.
            gdds[j,:] = gridread['dds']
            
            j=j+1

        gres = glis + gdims*gdds
        # Now we convert the glis, which were left edges (floats), to indices 
        # from the domain left edge.  Then we do a bunch of fixing now that we
        # know the extent of all the grids. 
        glis = np.round((glis - self.parameter_file.domain_left_edge)/gdds).astype('int')
        new_dre = np.max(gres,axis=0)
        self.parameter_file.domain_right_edge = np.round(new_dre, decimals=6)
        self.parameter_file.domain_width = \
                (self.parameter_file.domain_right_edge - 
                 self.parameter_file.domain_left_edge)
        self.parameter_file.domain_center = \
                0.5*(self.parameter_file.domain_left_edge + 
                     self.parameter_file.domain_right_edge)
        self.parameter_file.domain_dimensions = \
                np.round(self.parameter_file.domain_width/gdds[0]).astype('int')

        # Need to reset the units in the parameter file based on the correct
        # domain left/right/dimensions.
        self.parameter_file._set_units()

        if self.parameter_file.dimensionality <= 2 :
            self.parameter_file.domain_dimensions[2] = np.int(1)
        if self.parameter_file.dimensionality == 1 :
            self.parameter_file.domain_dimensions[1] = np.int(1)
        for i in range(levels.shape[0]):
            self.grids[i] = self.grid(i,self,levels[i],
                                      glis[i],
                                      gdims[i])
            dx = (self.parameter_file.domain_right_edge-
                  self.parameter_file.domain_left_edge)/self.parameter_file.domain_dimensions
            dx = dx/self.parameter_file.refine_by**(levels[i])
            dxs.append(dx)
        
        dx = np.array(dxs)
        self.grid_left_edge = np.round(self.parameter_file.domain_left_edge + dx*glis, decimals=6)
        self.grid_dimensions = gdims.astype("int32")
        self.grid_right_edge = np.round(self.grid_left_edge + dx*self.grid_dimensions, decimals=6)
        self.grid_particle_count = np.zeros([self.num_grids, 1], dtype='int64')

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()

        for g in self.grids:
            g.Children = self._get_grid_children(g)
            for g1 in g.Children:
                g1.Parent.append(g)
        self.max_level = self.grid_levels.max()

    def _get_grid_children(self, grid):
        mask = np.zeros(self.num_grids, dtype='bool')
        grids, grid_ind = self.get_box_grids(grid.LeftEdge, grid.RightEdge)
        mask[grid_ind] = True
        return [g for g in self.grids[mask] if g.Level == grid.Level + 1]

class AthenaStaticOutput(StaticOutput):
    _hierarchy_class = AthenaHierarchy
    _fieldinfo_fallback = AthenaFieldInfo
    _fieldinfo_known = KnownAthenaFields
    _data_style = "athena"

    def __init__(self, filename, data_style='athena',
                 storage_filename=None, parameters={}):
        self.specified_parameters = parameters
        StaticOutput.__init__(self, filename, data_style)
        self.filename = filename
        if storage_filename is None:
            storage_filename = '%s.yt' % filename.split('/')[-1]
        self.storage_filename = storage_filename

        # Unfortunately we now have to mandate that the hierarchy gets 
        # instantiated so that we can make sure we have the correct left 
        # and right domain edges.
        self.h

    def _set_units(self):
        """
        Generates the conversion to various physical _units based on the parameter file
        """
        self.units = {}
        self.time_units = {}
        if len(self.parameters) == 0:
            self._parse_parameter_file()
        self.conversion_factors = defaultdict(lambda: 1.0)    
        if self.specified_parameters.has_key("LengthUnits") :
            self._setup_getunits_units()
        else :
            self._setup_nounits_units()
        self.parameters["Time"] = self.conversion_factors["Time"]
        self.time_units['1'] = 1
        self.units['1'] = 1.0
        self.units['unitary'] = 1.0 / (self.domain_right_edge - self.domain_left_edge).max()
        for unit in sec_conversion.keys():
            self.time_units[unit] = self.conversion_factors["Time"] / sec_conversion[unit]
                        
    def _setup_getunits_units(self) :
        box_proper = 3.24077e-25 * self.specified_parameters["LengthUnits"]
        self.units['aye']  = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] * box_proper
        if self.specified_parameters.has_key("TimeUnits"):
            self.conversion_factors["Time"] = self.specified_parameters["TimeUnits"]
        else :
            self.conversion_factors["Time"] = 1.0
        if self.specified_parameters.has_key("DensityUnits"):
            self.conversion_factors["Density"] = self.specified_parameters["DensityUnits"]
        else :
            self.conversion_factors["Density"] = 1.0
        self.conversion_factors["Mass"] = self.conversion_factors["Density"]*self.units["cm"]**3
        for a in 'xyz':
            self.conversion_factors["%s-velocity" % (a)] = self.units["cm"]/self.conversion_factors["Time"]
                                            
    def _setup_nounits_units(self):
        self.conversion_factors["Time"] = 1.0
        self.conversion_factors["Density"] = 1.0
        self.conversion_factors["Mass"] = 1.0
        for a in 'xyz':
            self.conversion_factors["%s-velocity" % (a)] = 1.0
        for unit in mpc_conversion.keys():
            self.units[unit] = mpc_conversion[unit] / mpc_conversion["cm"]
        
    def _parse_parameter_file(self):
        self._handle = open(self.parameter_filename, "rb")
        # Read the start of a grid to get simulation parameters.
        grid = {}
        grid['read_field'] = None
        line = self._handle.readline()
        while grid['read_field'] is None:
            parse_line(line, grid)
            if "SCALAR" in line.strip().split():
                break
            if "VECTOR" in line.strip().split():
                break
            if 'TABLE' in line.strip().split():
                break
            if len(line) == 0: break
            line = self._handle.readline()

        self.domain_left_edge = grid['left_edge']
        mylog.info("Temporarily setting domain_right_edge = -domain_left_edge."+
                  " This will be corrected automatically if it is not the case.")
        self.domain_right_edge = -self.domain_left_edge
        self.domain_width = self.domain_right_edge-self.domain_left_edge
        self.domain_dimensions = np.round(self.domain_width/grid['dds']).astype('int32')
        refine_by = None
        if refine_by is None: refine_by = 2
        self.refine_by = refine_by
        dimensionality = 3
        if grid['dimensions'][2] == 1 :
            dimensionality = 2
        if grid['dimensions'][1] == 1 :
            dimensionality = 1
        if dimensionality <= 2 : self.domain_dimensions[2] = np.int32(1)
        if dimensionality == 1 : self.domain_dimensions[1] = np.int32(1)
        self.dimensionality = dimensionality
        self.current_time = grid["time"]
        self.unique_identifier = self.parameter_filename.__hash__()
        self.cosmological_simulation = False
        self.num_ghost_zones = 0
        self.field_ordering = 'fortran'
        self.boundary_conditions = [1]*6
        if 'periodicity' in self.specified_parameters:
            self.periodicity = ensure_tuple(self.specified_parameters['periodicity'])
        else:
            self.periodicity = (True,)*self.dimensionality

        dname = self.parameter_filename
        gridlistread = glob.glob('id*/%s-id*%s' % (dname[4:-9],dname[-9:] ))
        if 'id0' in dname :
            gridlistread += glob.glob('id*/lev*/%s*-lev*%s' % (dname[4:-9],dname[-9:]))
        else :
            gridlistread += glob.glob('lev*/%s*-lev*%s' % (dname[:-9],dname[-9:]))
        self.nvtk = len(gridlistread)+1 

        self.current_redshift = self.omega_lambda = self.omega_matter = \
            self.hubble_constant = self.cosmological_simulation = 0.0
        self.parameters['Time'] = self.current_time # Hardcode time conversion for now.
        self.parameters["HydroMethod"] = 0 # Hardcode for now until field staggering is supported.
        if self.specified_parameters.has_key("gamma") :
            self.parameters["Gamma"] = self.specified_parameters["gamma"]
        else :
            self.parameters["Gamma"] = 5./3. 
        self._handle.close()


    @classmethod
    def _is_valid(self, *args, **kwargs):
        try:
            if 'vtk' in args[0]:
                return True
        except:
            pass
        return False

    def __repr__(self):
        return self.basename.rsplit(".", 1)[0]




"""
libconfig (http://www.hyperrealm.com/libconfig/) wrapper in Cython.  Exposes a
minimum of the API.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from libc.stdio cimport fopen, fclose, FILE

cdef extern from "libconfig.h":
    cdef struct config_t:
        pass
    cdef struct config_setting_t:
        pass

    enum: CONFIG_TRUE
    enum: CONFIG_FALSE
    enum: CONFIG_TYPE_INT
    enum: CONFIG_TYPE_INT64
    enum: CONFIG_TYPE_FLOAT
    enum: CONFIG_TYPE_STRING
    enum: CONFIG_TYPE_BOOL
    enum: CONFIG_TYPE_ARRAY
    enum: CONFIG_TYPE_LIST
    enum: CONFIG_TYPE_GROUP

    void config_init (config_t *config)
    void config_destroy (config_t *config)
    int config_read (config_t * config, FILE * stream)
    int config_read_file (config_t * config, char * filename)
    int config_read_string (config_t * config, char * str)
    void config_write (config_t * config, FILE * stream)
    int config_write_file (config_t * config, char * filename)

    char * config_error_text (config_t * config)
    char * config_error_file (config_t * config)
    int config_error_line (config_t * config)

    config_setting_t * config_lookup (config_t * config, char * path)
    int config_lookup_int (config_t * config, char * path, int * value)
    int config_lookup_int64 (config_t * config, char * path, long long * value)
    int config_lookup_float (config_t * config, char * path, double * value)
    int config_lookup_bool (config_t * config, char * path, int * value)
    int config_lookup_string (config_t * config, char * path, char ** value)

    char * config_setting_name (config_setting_t * setting)

    int config_setting_get_int (config_setting_t * setting)
    long long config_setting_get_int64 (config_setting_t * setting)
    double config_setting_get_float (config_setting_t * setting)
    int config_setting_get_bool (config_setting_t * setting)
    char * config_setting_get_string (config_setting_t * setting)

    int config_setting_type (config_setting_t * setting)

    int config_setting_length (config_setting_t * setting)
    config_setting_t * config_setting_get_elem (
        config_setting_t * setting, unsigned int idx)

cdef class libconfigSetting:
    cdef config_setting_t *setting

    def keys(self):
        cdef int tt = config_setting_type(self.setting)
        cdef char *setting_name
        cdef config_setting_t *sub_setting
        if tt != CONFIG_TYPE_GROUP: return []
        nelem = config_setting_length(self.setting)
        tr = []
        for i in xrange(nelem):
            sub_setting = config_setting_get_elem(self.setting, i)
            setting_name = config_setting_name(sub_setting)
            tr.append(setting_name)
        return tr

cdef class libconfigConfiguration:
    cdef config_t cfg
    def __cinit__(self):
        config_init(&self.cfg)

    def __dealloc__(self):
        config_destroy(&self.cfg)

    def read_file(self, char *fn):
        cdef int status, error_line
        cdef char *error_text, *error_file
        status = config_read_file(&self.cfg, fn)
        if status == CONFIG_FALSE:
            error_text = config_error_text(&self.cfg)
            error_file = config_error_file(&self.cfg)
            error_line = config_error_line(&self.cfg)
            if error_file != NULL:
                print "Error encounters in %s at line %s" % (
                        error_file, error_line)
            print error_text

    def read_string(self, char *s):
        config_read_string(&self.cfg, s)

    def __getitem__(self, key):
        cdef config_setting_t *setting = config_lookup(&self.cfg, key)
        ms = libconfigSetting()
        ms.setting = setting
        rv = self.get_setting(ms)
        if rv == None: raise KeyError(key)
        return rv

    def get_setting(self, libconfigSetting ms):
        cdef config_setting_t *setting = ms.setting
        if setting == NULL:
            return None
        cdef int tt = config_setting_type(setting)
        cdef int kv_int
        cdef long long kv_int64
        cdef double kv_float
        cdef int kv_bool
        cdef char* kv_string
        cdef int nelem
        cdef config_setting_t *sub_setting
        if tt == CONFIG_TYPE_INT:
            kv_int = config_setting_get_int(setting)
            return kv_int
        elif tt == CONFIG_TYPE_INT64:
            kv_int64 = config_setting_get_int64(setting)
            return kv_int64
        elif tt == CONFIG_TYPE_FLOAT:
            kv_float = config_setting_get_float(setting)
            return kv_float
        elif tt == CONFIG_TYPE_BOOL:
            kv_bool = config_setting_get_bool(setting)
            return kv_bool
        elif tt == CONFIG_TYPE_STRING:
            kv_string = config_setting_get_string(setting)
            return kv_string
        elif tt == CONFIG_TYPE_ARRAY or tt == CONFIG_TYPE_LIST:
            nelem = config_setting_length(setting)
            tr = []
            for i in xrange(nelem):
                sub_ms = libconfigSetting()
                sub_setting = config_setting_get_elem(setting, i)
                sub_ms.setting = sub_setting
                tr.append(self.get_setting(sub_ms))
            return tr
        elif tt == CONFIG_TYPE_GROUP:
            group = libconfigSetting()
            group.setting = setting
            return group
        return None



"""
StarAnalysis - Functions to analyze stars.

Author: Stephen Skory <s@skory.us>
Affiliation: UC San Diego / CASS
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Stephen Skory (and others).  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import h5py
import math, itertools

from yt.funcs import *
from yt.utilities.cosmology import \
    Cosmology, \
    EnzoCosmology

YEAR = 3.155693e7 # sec / year
LIGHT = 2.997925e10 # cm / s

class StarFormationRate(object):
    r"""Calculates the star formation rate for a given population of
    star particles.
    
    Parameters
    ----------
    pf : EnzoStaticOutput object
    data_source : AMRRegion object, optional
        The region from which stars are extracted for analysis. If this
        is not supplied, the next three must be, otherwise the next
        three do not need to be specified.
    star_mass : Ordered array or list of floats
        The mass of the stars to be analyzed in units of Msun.
    star_creation_time : Ordered array or list of floats
        The creation time for the stars in code units.
    volume : Float
        The volume of the region for the specified list of stars.
    bins : Integer
        The number of time bins used for binning the stars. Default = 300.
    
    Examples
    --------
    
    >>> pf = load("RedshiftOutput0000")
    >>> sp = pf.h.sphere([0.5,0.5,0.5], [.1])
    >>> sfr = StarFormationRate(pf, sp)
    """
    def __init__(self, pf, data_source=None, star_mass=None,
            star_creation_time=None, volume=None, bins=300):
        self._pf = pf
        self._data_source = data_source
        self.star_mass = np.array(star_mass)
        self.star_creation_time = np.array(star_creation_time)
        self.volume = volume
        self.bin_count = bins
        # Check to make sure we have the right set of informations.
        if data_source is None:
            if self.star_mass is None or self.star_creation_time is None or \
            self.volume is None:
                mylog.error(
                """
                If data_source is not provided, all of these paramters need to be set:
                star_mass (array, Msun),
                star_creation_time (array, code units),
                volume (float, Mpc**3).
                """)
                return None
            self.mode = 'provided'
        else:
            self.mode = 'data_source'
        # Set up for time conversion.
        self.cosm = EnzoCosmology(HubbleConstantNow = 
             (100.0 * self._pf.hubble_constant),
             OmegaMatterNow = self._pf.omega_matter,
             OmegaLambdaNow = self._pf.omega_lambda,
             InitialRedshift = self._pf['CosmologyInitialRedshift'])
        # Find the time right now.
        self.time_now = self.cosm.ComputeTimeFromRedshift(
            self._pf.current_redshift) # seconds
        # Build the distribution.
        self.build_dist()
        # Attach some convenience arrays.
        self.attach_arrays()

    def build_dist(self):
        """
        Build the data for plotting.
        """
        # Pick out the stars.
        if self.mode == 'data_source':
            ct = self._data_source["creation_time"]
            ct_stars = ct[ct > 0]
            mass_stars = self._data_source["ParticleMassMsun"][ct > 0]
        elif self.mode == 'provided':
            ct_stars = self.star_creation_time
            mass_stars = self.star_mass
        # Find the oldest stars in units of code time.
        tmin= min(ct_stars)
        # Multiply the end to prevent numerical issues.
        self.time_bins = np.linspace(tmin*0.99, self._pf.current_time,
            num = self.bin_count + 1)
        # Figure out which bins the stars go into.
        inds = np.digitize(ct_stars, self.time_bins) - 1
        # Sum up the stars created in each time bin.
        self.mass_bins = np.zeros(self.bin_count + 1, dtype='float64')
        for index in np.unique(inds):
            self.mass_bins[index] += sum(mass_stars[inds == index])
        # Calculate the cumulative mass sum over time by forward adding.
        self.cum_mass_bins = self.mass_bins.copy()
        for index in xrange(self.bin_count):
            self.cum_mass_bins[index+1] += self.cum_mass_bins[index]
        # We will want the time taken between bins.
        self.time_bins_dt = self.time_bins[1:] - self.time_bins[:-1]
    
    def attach_arrays(self):
        """
        Attach convenience arrays to the class for easy access.
        """
        if self.mode == 'data_source':
            try:
                vol = self._data_source.volume('mpc')
            except AttributeError:
                # If we're here, this is probably a HOPHalo object, and we
                # can get the volume this way.
                ds = self._data_source.get_sphere()
                vol = ds.volume('mpc')
        elif self.mode == 'provided':
            vol = self.volume
        tc = self._pf["Time"]
        self.time = []
        self.lookback_time = []
        self.redshift = []
        self.Msol_yr = []
        self.Msol_yr_vol = []
        self.Msol = []
        self.Msol_cumulative = []
        # Use the center of the time_bin, not the left edge.
        for i, time in enumerate((self.time_bins[1:] + self.time_bins[:-1])/2.):
            self.time.append(time * tc / YEAR)
            self.lookback_time.append((self.time_now - time * tc)/YEAR)
            self.redshift.append(self.cosm.ComputeRedshiftFromTime(time * tc))
            self.Msol_yr.append(self.mass_bins[i] / \
                (self.time_bins_dt[i] * tc / YEAR))
            self.Msol_yr_vol.append(self.mass_bins[i] / \
                (self.time_bins_dt[i] * tc / YEAR) / vol)
            self.Msol.append(self.mass_bins[i])
            self.Msol_cumulative.append(self.cum_mass_bins[i])
        self.time = np.array(self.time)
        self.lookback_time = np.array(self.lookback_time)
        self.redshift = np.array(self.redshift)
        self.Msol_yr = np.array(self.Msol_yr)
        self.Msol_yr_vol = np.array(self.Msol_yr_vol)
        self.Msol = np.array(self.Msol)
        self.Msol_cumulative = np.array(self.Msol_cumulative)
    
    def write_out(self, name="StarFormationRate.out"):
        r"""Write out the star analysis to a text file *name*. The columns are in
        order.

        The columns in the output file are:
           1. Time (yrs)
           2. Look-back time (yrs)
           3. Redshift
           4. Star formation rate in this bin per year (Msol/yr)
           5. Star formation rate in this bin per year per Mpc**3 (Msol/yr/Mpc**3)
           6. Stars formed in this time bin (Msol)
           7. Cumulative stars formed up to this time bin (Msol)
        
        Parameters
        ----------
        name : String
            The name of the file to write to. Default = StarFormationRate.out.
        
        Examples
        --------
        >>> sfr.write_out("stars-SFR.out")
        """
        fp = open(name, "w")
        fp.write("#time\tlookback\tredshift\tMsol/yr\tMsol/yr/Mpc3\tMsol\tcumMsol\t\n")
        for i, time in enumerate(self.time):
            line = "%1.5e %1.5e %1.5e %1.5e %1.5e %1.5e %1.5e\n" % \
            (time, # Time
            self.lookback_time[i], # Lookback time
            self.redshift[i], # Redshift
            self.Msol_yr[i], # Msol/yr
            self.Msol_yr_vol[i], # Msol/yr/vol
            self.Msol[i], # Msol in bin
            self.Msol_cumulative[i]) # cumulative
            fp.write(line)
        fp.close()

### Begin Synthetic Spectrum Stuff. ####

CHABRIER = {
"Z0001" : "bc2003_hr_m22_chab_ssp.ised.h5", #/* 0.5% */
"Z0004" : "bc2003_hr_m32_chab_ssp.ised.h5", #/* 2% */
"Z004" : "bc2003_hr_m42_chab_ssp.ised.h5", #/* 20% */
"Z008" : "bc2003_hr_m52_chab_ssp.ised.h5", #/* 40% */
"Z02" : "bc2003_hr_m62_chab_ssp.ised.h5", #/* solar; 0.02 */
"Z05" : "bc2003_hr_m72_chab_ssp.ised.h5" #/* 250% */
}

SALPETER = {
"Z0001" : "bc2003_hr_m22_salp_ssp.ised.h5", #/* 0.5% */
"Z0004" : "bc2003_hr_m32_salp_ssp.ised.h5", #/* 2% */
"Z004" : "bc2003_hr_m42_salp_ssp.ised.h5", #/* 20% */
"Z008" : "bc2003_hr_m52_salp_ssp.ised.h5", #/* 40% */
"Z02" : "bc2003_hr_m62_salp_ssp.ised.h5", #/* solar; 0.02 */
"Z05" : "bc2003_hr_m72_salp_ssp.ised.h5" #/* 250% */
}

Zsun = 0.02

#/* dividing line of metallicity; linear in log(Z/Zsun) */
METAL1 = 0.01  # /* in units of Z/Zsun */
METAL2 = 0.0632
METAL3 = 0.2828
METAL4 = 0.6325
METAL5 = 1.5811
METALS = np.array([METAL1, METAL2, METAL3, METAL4, METAL5])

# Translate METALS array digitize to the table dicts
MtoD = np.array(["Z0001", "Z0004", "Z004", "Z008", "Z02",  "Z05"])

"""
This spectrum code is based on code from Ken Nagamine, converted from C to Python.
I've also reversed the order of elements in the flux arrays to be in C-ordering,
for faster memory access.
"""

class SpectrumBuilder(object):
    r"""Initialize the data to build a summed flux spectrum for a
    collection of stars using the models of Bruzual & Charlot (2003).
    This function loads the necessary data tables into memory and
    must be called before analyzing any star particles.
    
    Parameters
    ----------
    pf : EnzoStaticOutput object
    bcdir : String
        Path to directory containing Bruzual & Charlot h5 fit files.
    model : String
        Choice of Initial Metalicity Function model, 'chabrier' or
        'salpeter'. Default = 'chabrier'.
    
    Examples
    --------
    >>> pf = load("RedshiftOutput0000")
    >>> spec = SpectrumBuilder(pf, "/home/user/bc/", model="salpeter")
    """
    def __init__(self, pf, bcdir="", model="chabrier", time_now=None):
        self._pf = pf
        self.bcdir = bcdir
        
        if model == "chabrier":
            self.model = CHABRIER
        elif model == "salpeter":
            self.model = SALPETER
        # Set up for time conversion.
        self.cosm = EnzoCosmology(HubbleConstantNow = 
             (100.0 * self._pf.hubble_constant),
             OmegaMatterNow = self._pf.omega_matter,
             OmegaLambdaNow = self._pf.omega_lambda,
             InitialRedshift = self._pf['CosmologyInitialRedshift'])
        # Find the time right now.
        
        if time_now is None:
            self.time_now = self.cosm.ComputeTimeFromRedshift(
                self._pf.current_redshift) # seconds
        else:
            self.time_now = time_now
        
        # Read the tables.
        self.read_bclib()

    def read_bclib(self):
        """
        Read in the age and wavelength bins, and the flux bins for each
        metallicity.
        """
        self.flux = {}
        for file in self.model:
            fname = self.bcdir + "/" + self.model[file]
            fp = h5py.File(fname, 'r')
            self.age = fp["agebins"][:] # 1D floats
            self.wavelength = fp["wavebins"][:] # 1D floats
            self.flux[file] = fp["flam"][:,:] # 2D floats, [agebin, wavebin]
            fp.close()
    
    def calculate_spectrum(self, data_source=None, star_mass=None,
            star_creation_time=None, star_metallicity_fraction=None,
            star_metallicity_constant=None, min_age=0.):
        r"""For the set of stars, calculate the collective spectrum.
        Attached to the output are several useful objects:
        final_spec: The collective spectrum in units of flux binned in wavelength.
        wavelength: The wavelength for the spectrum bins, in Angstroms.
        total_mass: Total mass of all the stars.
        avg_mass: Average mass of all the stars.
        avg_metal: Average metallicity of all the stars.
        
        Parameters
        ----------
        data_source : AMRRegion object, optional
            The region from which stars are extracted for analysis. If this is
            not specified, the next three parameters must be supplied.
        star_mass : Array or list of floats
            An array of star masses in Msun units.
        star_creation_time : Array or list of floats
            An array of star creation times in code units.
        star_metallicity_fraction : Array or list of floats
            An array of star metallicity fractions, in code
                units (which is not Z/Zsun, rather just Z).
        star_metallicity_constant : Float
            If desired, override the star
            metallicity fraction of all the stars to the given value.
        min_age : Float
            Removes young stars younger than this number (in years)
            from the spectrum. Default: 0 (all stars).
        
        Examples
        --------
        >>> sp = pf.h.sphere([0.5,0.5,0.5], [.1])
        >>> spec.calculate_spectrum(data_source=sp, min_age = 1.e6)
        """
        # Initialize values
        self.final_spec = np.zeros(self.wavelength.size, dtype='float64')
        self._data_source = data_source
        if iterable(star_mass):
            self.star_mass = star_mass
        else:
            self.star_mass = [star_mass]
        if iterable(star_creation_time):
            self.star_creation_time = star_creation_time
        else:
            self.star_creation_time = [star_creation_time]
        if iterable(star_metallicity_fraction):
            self.star_metal = star_metallicity_fraction
        else:
            self.star_metal = [star_metallicity_fraction]
        self.min_age = min_age
        
        # Check to make sure we have the right set of data.
        if data_source is None:
            if self.star_mass is None or self.star_creation_time is None or \
            (star_metallicity_fraction is None and star_metallicity_constant is None):
                mylog.error(
                """
                If data_source is not provided, all of these paramters need to be set:
                star_mass (array, Msun),
                star_creation_time (array, code units),
                And one of:
                star_metallicity_fraction (array, code units).
                --OR--
                star_metallicity_constant (float, code units).
                """)
                return None
            if star_metallicity_constant is not None:
                self.star_metal = np.ones(self.star_mass.size, dtype='float64') * \
                    star_metallicity_constant
            if star_metallicity_fraction is not None:
                self.star_metal = star_metallicity_fraction
        else:
            # Get the data we need.
            ct = self._data_source["creation_time"]
            self.star_creation_time = ct[ct > 0]
            self.star_mass = self._data_source["ParticleMassMsun"][ct > 0]
            if star_metallicity_constant is not None:
                self.star_metal = np.ones(self.star_mass.size, dtype='float64') * \
                    star_metallicity_constant
            else:
                self.star_metal = self._data_source["metallicity_fraction"][ct > 0]
        # Fix metallicity to units of Zsun.
        self.star_metal /= Zsun
        # Age of star in years.
        dt = (self.time_now - self.star_creation_time * self._pf['Time']) / YEAR
        dt = np.maximum(dt, 0.0)
        # Remove young stars
        sub = dt >= self.min_age
        if len(sub) == 0: return
        self.star_metal = self.star_metal[sub]
        dt = dt[sub]
        self.star_creation_time = self.star_creation_time[sub]
        # Figure out which METALS bin the star goes into.
        Mindex = np.digitize(self.star_metal, METALS)
        # Replace the indices with strings.
        Mname = MtoD[Mindex]
        # Figure out which age bin this star goes into.
        Aindex = np.digitize(dt, self.age)
        # Ratios used for the interpolation.
        ratio1 = (dt - self.age[Aindex-1]) / (self.age[Aindex] - self.age[Aindex-1])
        ratio2 = (self.age[Aindex] - dt) / (self.age[Aindex] - self.age[Aindex-1])
        # Sort the stars by metallicity and then by age, which should reduce
        # memory access time by a little bit in the loop.
        indexes = np.arange(self.star_metal.size)
        sort = np.asarray([indexes[i] for i in np.lexsort([indexes, Aindex, Mname])])
        Mname = Mname[sort]
        Aindex = Aindex[sort]
        ratio1 = ratio1[sort]
        ratio2 = ratio2[sort]
        self.star_mass = self.star_mass[sort]
        self.star_creation_time = self.star_creation_time[sort]
        self.star_metal = self.star_metal[sort]
        
        # Interpolate the flux for each star, adding to the total by weight.
        pbar = get_pbar("Calculating fluxes",len(self.star_mass))
        for i,star in enumerate(itertools.izip(Mname, Aindex, ratio1, ratio2, self.star_mass)):
            # Pick the right age bin for the right flux array.
            flux = self.flux[star[0]][star[1],:]
            # Get the one just before the one above.
            flux_1 = self.flux[star[0]][star[1]-1,:]
            # interpolate in log(flux), linear in time.
            int_flux = star[3] * np.log10(flux_1) + star[2] * np.log10(flux)
            # Add this flux to the total, weighted by mass.
            self.final_spec += np.power(10., int_flux) * star[4]
            pbar.update(i)
        pbar.finish()    
        
        # Normalize.
        self.total_mass = np.sum(self.star_mass)
        self.avg_mass = np.mean(self.star_mass)
        tot_metal = sum(self.star_metal * self.star_mass)
        self.avg_metal = math.log10(tot_metal / self.total_mass / Zsun)

        # Below is an attempt to do the loop using vectors and matrices,
        # however it doesn't appear to be much faster, probably due to all
        # the gymnastics that have to be done to do element-by-element
        # multiplication for matricies.
        # I'm keeping it in here in case I come up with a more
        # elegant way that actually is faster.
#         for metal_name in MtoD:
#             # Pick out our stars in this metallicity bin.
#             select = (Mname == metal_name)
#             A = Aindex[select]
#             if A.size == 0: continue
#             r1 = ratio1[select]
#             r2 = ratio2[select]
#             sm = self.star_mass[select]
#             # From the flux array for this metal, and our selection, build
#             # a new flux array just for the ages of these stars, in the 
#             # same order as the selection of stars.
#             this_flux = np.matrix(self.flux[metal_name][A])
#             # Make one for the last time step for each star in the same fashion
#             # as above.
#             this_flux_1 = np.matrix(self.flux[metal_name][A-1])
#             # This is kind of messy, but we're going to multiply this_fluxes
#             # by the appropriate ratios and add it together to do the 
#             # interpolation in log(flux) and linear in time.
#             print r1.size
#             r1 = np.matrix(r1.tolist()*self.wavelength.size).reshape(self.wavelength.size,r1.size).T
#             r2 = np.matrix(r2.tolist()*self.wavelength.size).reshape(self.wavelength.size,r2.size).T
#             print this_flux_1.shape, r1.shape
#             int_flux = np.multiply(np.log10(this_flux_1),r1) \
#                 + np.multiply(np.log10(this_flux),r2)
#             # Weight the fluxes by mass.
#             sm = np.matrix(sm.tolist()*self.wavelength.size).reshape(self.wavelength.size,sm.size).T
#             int_flux = np.multiply(np.power(10., int_flux), sm)
#             # Sum along the columns, converting back to an array, adding
#             # to the full spectrum.
#             self.final_spec += np.array(int_flux.sum(axis=0))[0,:]

    
    def write_out(self, name="sum_flux.out"):
        r"""Write out the summed flux to a file. 

        The output file from this function has two columns: Wavelength
        (Angstrom) and Flux (Luminosity per unit wavelength, L_sun Ang^-1,
        L_sun = 3.826 * 10^33 ergs s^-1.).

        Parameters
        ----------
        name : String
            Name of file to write to. Default = "sum_flux.out"
        
        Examples
        --------
        >>> spec.write_out("spec.out")
        """
        fp = open(name, 'w')
        for i, wave in enumerate(self.wavelength):
            fp.write("%1.5e\t%1.5e\n" % (wave, self.final_spec[i]))
        fp.close()

    def write_out_SED(self, name="sum_SED.out", flux_norm=5200.):
        r"""Write out the summed SED to a file. The file has two columns:
        1) Wavelength (Angstrom)
        2) Relative flux normalized to the flux at *flux_norm*.
        It also will attach to the SpectrumBuilder object 
        an array *f_nu* which is the normalized flux,
        identical to the disk output.
        
        Parameters
        ----------
        name : String
            Name of file to write to. Default = "sum_SED.out"
        flux_norm : Float
            Wavelength of the flux to normalize the distribution against.
            Default = 5200 Ang.
        
        Examples
        --------
        >>> spec.write_out_SED(name = "SED.out", flux_norm = 6000.)
        """
        # find the f_nu closest to flux_norm
        fn_wavelength = np.argmin(abs(self.wavelength - flux_norm))
        f_nu = self.final_spec * np.power(self.wavelength, 2.) / LIGHT
        # Normalize f_nu
        self.f_nu = f_nu / f_nu[fn_wavelength]
        # Write out.
        fp = open(name, 'w')
        for i, wave in enumerate(self.wavelength):
            fp.write("%1.5e\t%1.5e\n" % (wave, self.f_nu[i]))
        fp.close()




"""
Answer Testing support for Enzo.

Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: Michigan State University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.testing import *
from yt.config import ytcfg
from yt.mods import *

from yt.utilities.answer_testing.framework import \
     AnswerTestingTest, \
     can_run_pf, \
     FieldValuesTest, \
     GridHierarchyTest, \
     GridValuesTest, \
     ProjectionValuesTest, \
     ParentageRelationshipsTest, \
     temp_cwd, \
     AssertWrapper

def requires_outputlog(path = ".", prefix = ""):
    def ffalse(func):
        return lambda: None
    def ftrue(func):
        @wraps(func)
        def fyielder(*args, **kwargs):
            with temp_cwd(path):
                for t in func(*args, **kwargs):
                    if isinstance(t, AnswerTestingTest):
                        t.prefix = prefix
                    yield t
        return fyielder
    if os.path.exists("OutputLog"):
        return ftrue
    with temp_cwd(path):
        if os.path.exists("OutputLog"):
            return ftrue
    return ffalse
     
def standard_small_simulation(pf_fn, fields):
    if not can_run_pf(pf_fn): return
    dso = [None]
    tolerance = ytcfg.getint("yt", "answer_testing_tolerance")
    bitwise = ytcfg.getboolean("yt", "answer_testing_bitwise")
    for field in fields:
        if bitwise:
            yield GridValuesTest(pf_fn, field)
        if 'particle' in field: continue
        for ds in dso:
            for axis in [0, 1, 2]:
                for weight_field in [None, "Density"]:
                    yield ProjectionValuesTest(
                        pf_fn, axis, field, weight_field,
                        ds, decimals=tolerance)
            yield FieldValuesTest(
                    pf_fn, field, ds, decimals=tolerance)
                    
class ShockTubeTest(object):
    def __init__(self, data_file, solution_file, fields, 
                 left_edges, right_edges, rtol, atol):
        self.solution_file = solution_file
        self.data_file = data_file
        self.fields = fields
        self.left_edges = left_edges
        self.right_edges = right_edges
        self.rtol = rtol
        self.atol = atol

    def __call__(self):
        # Read in the pf
        pf = load(self.data_file)  
        exact = self.get_analytical_solution() 

        ad = pf.h.all_data()
        position = ad['x']
        for k in self.fields:
            field = ad[k]
            for xmin, xmax in zip(self.left_edges, self.right_edges):
                mask = (position >= xmin)*(position <= xmax)
                exact_field = np.interp(position[mask], exact['pos'], exact[k]) 
                myname = "ShockTubeTest_%s" % k
                # yield test vs analytical solution 
                yield AssertWrapper(myname, assert_allclose, field[mask], 
                                    exact_field, self.rtol, self.atol)

    def get_analytical_solution(self):
        # Reads in from file 
        pos, dens, vel, pres, inte = \
                np.loadtxt(self.solution_file, unpack=True)
        exact = {}
        exact['pos'] = pos
        exact['Density'] = dens
        exact['x-velocity'] = vel
        exact['Pressure'] = pres
        exact['ThermalEnergy'] = inte
        return exact



"""
Import the components of the volume rendering extension

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from yt.funcs import *
import h5py

from yt.utilities.lib import PartitionedGrid, ProtoPrism, GridFace, \
    grid_points_in_volume, find_grids_in_inclined_box
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface, parallel_root_only

# HISTORICAL NOTE OF SOME IMPORT:
#   The homogenized brick class should and will be removed.  The more general,
#   and less fancy-parallel, HomogenizedVolume is the way to go.  The HBC is a
#   complicated mechanism that was designed for back when we were going to be
#   passing bricks all around between processors.

class HomogenizedVolume(ParallelAnalysisInterface):
    bricks = None
    def __init__(self, fields = "Density", source = None, pf = None,
                 log_fields = None, no_ghost = False,
                 max_level = 48):
        # Typically, initialized as hanging off a hierarchy.  But, not always.
        ParallelAnalysisInterface.__init__(self)
        self.no_ghost = no_ghost
        if pf is not None: self.pf = pf
        if source is None: source = self.pf.h.all_data()
        self.source = source
        self.fields = ensure_list(fields)
        if log_fields is not None:
            log_fields = ensure_list(log_fields)
        else:
            log_fields = [self.pf.field_info[field].take_log
                         for field in self.fields]
        self.max_level = max_level
        self.log_fields = log_fields

    def traverse(self, back_point, front_point, image):
        mylog.info("Traversing %s bricks between %s and %s",
                   len(self.bricks), back_point, front_point)
        if self.bricks is None: self.initialize_source()
        vec = front_point - back_point
        dist = np.minimum(
             np.sum((self.brick_left_edges - back_point) * vec, axis=1),
             np.sum((self.brick_right_edges - back_point) * vec, axis=1))
        ind = np.argsort(dist)
        for b in self.bricks[ind]:
            #print b.LeftEdge, b.RightEdge
            yield b

    def _partition_grid(self, grid):

        # This is not super efficient, as it re-fills the regions once for each
        # field.
        vcds = []
        for field, log_field in zip(self.fields, self.log_fields):
            vcd = grid.get_vertex_centered_data(field, no_ghost = self.no_ghost)
            vcd = vcd.astype("float64")
            if log_field: vcd = np.log10(vcd)
            vcds.append(vcd)

        GF = GridFaces(grid.Children + [grid])
        PP = ProtoPrism(grid.id, grid.LeftEdge, grid.RightEdge, GF)

        pgs = []
        cm = grid.child_mask.copy()
        if grid.Level > self.max_level:
            return pgs
        elif grid.Level == self.max_level:
            cm[:] = 1
        for P in PP.sweep(0):
            sl = P.get_brick(grid.LeftEdge, grid.dds, cm)
            if len(sl) == 0: continue
            dd = [d[sl[0][0]:sl[0][1]+1,
                    sl[1][0]:sl[1][1]+1,
                    sl[2][0]:sl[2][1]+1].copy() for d in vcds]
            pgs.append(PartitionedGrid(grid.id, dd,
                        P.LeftEdge, P.RightEdge, sl[-1]))
        return pgs

    def initialize_source(self, source = None):
        if self.bricks is not None and source is not None:
            raise NotImplementedError("Sorry, dynamic shuffling of bricks is" + 
                                      " not yet supported")
        if self.bricks is not None and source is None: return
        bricks = []
        self.comm.preload(self.source._grids, self.fields, self.pf.h.io)
        pbar = get_pbar("Partitioning ", len(self.source._grids))
        for i, g in enumerate(self.source._grids):
            pbar.update(i)
            bricks += self._partition_grid(g)
        pbar.finish()
        self.initialize_bricks(bricks)

    def initialize_bricks(self, bricks):
        NB = len(bricks)
        # Now we set up our (local for now) hierarchy.  Note that to calculate
        # intersection, we only need to do the left edge & right edge.
        #
        # We're going to double up a little bit here in memory.
        self.brick_left_edges = np.zeros( (NB, 3), dtype='float64')
        self.brick_right_edges = np.zeros( (NB, 3), dtype='float64')
        self.brick_parents = np.zeros( NB, dtype='int64')
        self.brick_dimensions = np.zeros( (NB, 3), dtype='int64')
        self.bricks = np.empty(len(bricks), dtype='object')
        for i,b in enumerate(bricks):
            self.brick_left_edges[i,:] = b.LeftEdge
            self.brick_right_edges[i,:] = b.RightEdge
            self.brick_parents[i] = b.parent_grid_id
            self.brick_dimensions[i,:] = b.my_data[0].shape
            self.bricks[i] = b
        # Vertex-centered means we subtract one from the shape
        self.brick_dimensions -= 1

    def reflect_across_boundaries(self):
        mylog.warning("Note that this doesn't fix ghost zones, so there may be artifacts at domain boundaries!")
        nb = []
        # Simplest, clearest iteration ...
        for i in [-1, 1]:
            for j in [-1, 1]:
                for k in [-1, 1]:
                    for b in self.bricks:
                        BB = np.array([b.LeftEdge * [i,j,k], b.RightEdge * [i,j,k]])
                        LE, RE = np.min(BB, axis=0), np.max(BB, axis=0)
                        nb.append(
                            PartitionedGrid(b.parent_grid_id, len(b.my_data), 
                                [md[::i,::j,::k].copy("C") for md in b.my_data],
                                LE, RE, np.array(b.my_data[0].shape) - 1))
        # Replace old bricks
        self.initialize_bricks(nb)

    def store_bricks(self, fn):
        import h5py, cPickle
        f = h5py.File(fn, "w")
        f.create_dataset("/left_edges", data=self.brick_left_edges)
        f.create_dataset("/right_edges", data=self.brick_right_edges)
        f.create_dataset("/parents", data=self.brick_parents)
        f.create_dataset("/dimensions", data=self.brick_dimensions)
        f.create_group("/bricks")
        for i,b in enumerate(self.bricks):
            f.create_group("/bricks/brick_%08i" % i)
            for fi,field in enumerate(self.fields):
                f.create_dataset("/bricks/brick_%08i/%s" % (i, field),
                                 data=b.my_data[fi])
        f.close()

    def load_bricks(self, fn):
        import h5py
        f = h5py.File(fn, "r")
        self.brick_left_edges = f["/left_edges"][:]
        self.brick_right_edges = f["/right_edges"][:]
        self.brick_parents = f["/parents"][:]
        self.brick_dimensions= f["/dimensions"][:]
        bricks = []
        for i,ds in enumerate(sorted(f["/bricks"])):
            td = [f["/bricks/%s/%s" % (ds, field)][:] for field in self.fields]
            bricks.append(PartitionedGrid(
                                self.brick_parents[i], len(td), td,
                                self.brick_left_edges[i,:],
                                self.brick_right_edges[i,:],
                                self.brick_dimensions[i,:],
                                ))
        self.bricks = np.array(bricks, dtype='object')
        f.close()

    def reset_cast(self):
        pass

class SingleBrickVolume(object):
    bricks = None
    def __init__(self, data_array):
        self.bricks = [PartitionedGrid(-1, 1, 
                       [data_array.astype("float64")],
                       np.zeros(3, dtype='float64'),
                       np.ones(3, dtype='float64'),
                       np.array(data_array.shape, dtype='int64')-1)]
        self.brick_dimensions = np.ones((1, 3), dtype='int64')*data_array.shape

    def initialize_source(self):
        pass

    def traverse(self, back, front, image):
        for b in self.bricks: yield b

    def reset_cast(self):
        pass

class GridFaces(object):
    def __init__(self, grids):
        self.faces = [ [], [], [] ]
        for grid in grids:
            for direction in range(3):
                self.faces[direction].append( GridFace(grid, direction, 1) )
                self.faces[direction].append( GridFace(grid, direction, 0) )
        for f in self.faces:
            f.sort(key = lambda a: a.coord)

    def __getitem__(self, item):
        return self.faces[item]

def export_partitioned_grids(grid_list, fn, int_type=np.int64, float_type=np.float64):
    f = h5py.File(fn, "w")
    pbar = get_pbar("Writing Grids", len(grid_list))
    nelem = sum((grid.my_data.size for grid in grid_list))
    ngrids = len(grid_list)
    group = f.create_group("/PGrids")
    left_edge = np.concatenate([[grid.LeftEdge,] for grid in grid_list])
    f.create_dataset("/PGrids/LeftEdges", data=left_edge, dtype=float_type); del left_edge
    right_edge = np.concatenate([[grid.RightEdge,] for grid in grid_list])
    f.create_dataset("/PGrids/RightEdges", data=right_edge, dtype=float_type); del right_edge
    dims = np.concatenate([[grid.my_data.shape[:],] for grid in grid_list])
    f.create_dataset("/PGrids/Dims", data=dims, dtype=int_type); del dims
    data = np.concatenate([grid.my_data.ravel() for grid in grid_list])
    f.create_dataset("/PGrids/Data", data=data, dtype=float_type); del data
    f.close()
    pbar.finish()

def import_partitioned_grids(fn, int_type=np.int64, float_type=np.float64):
    f = h5py.File(fn, "r")
    n_groups = len(f)
    grid_list = []
    dims = f["/PGrids/Dims"][:].astype(int_type)
    left_edges = f["/PGrids/LeftEdges"][:].astype(float_type)
    right_edges = f["/PGrids/RightEdges"][:].astype(float_type)
    data = f["/PGrids/Data"][:].astype(float_type)
    pbar = get_pbar("Reading Grids", dims.shape[0])
    curpos = 0
    for i in xrange(dims.shape[0]):
        gd = dims[i,:]
        gle, gre = left_edges[i,:], right_edges[i,:]
        gdata = data[curpos:curpos+gd.prod()].reshape(gd)
        # Vertex -> Grid, so we -1 from dims in this
        grid_list.append(PartitionedGrid(-1, gdata, gle, gre, gd - 1))
        curpos += gd.prod()
        pbar.update(i)
    pbar.finish()
    f.close()
    return np.array(grid_list, dtype='object')



"""
HaloProfiler class and member functions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import gc
import numpy as np
import os
import h5py
import types

from yt.funcs import *
from yt.utilities.math_utils import periodic_dist

from yt.convenience import \
    load
from yt.data_objects.profiles import \
    BinnedProfile1D, EmptyProfileData
from yt.analysis_modules.halo_finding.api import *
from .halo_filters import \
    VirialFilter
from .centering_methods import \
    centering_registry
from yt.data_objects.field_info_container import \
    add_field
from yt.data_objects.static_output import \
    StaticOutput

from yt.utilities.exceptions import \
    YTException
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface, \
    parallel_blocking_call, \
    parallel_root_only, \
    parallel_objects
from yt.visualization.fixed_resolution import \
    FixedResolutionBuffer
from yt.visualization.image_writer import write_image

PROFILE_RADIUS_THRESHOLD = 2

class HaloProfiler(ParallelAnalysisInterface):
    r"""Initialize a Halo Profiler object.

    In order to run the halo profiler, the Halo Profiler object must be
    instantiated. At the minimum, the path to a parameter file
    must be provided as the first term.

    Parameters
    ----------

    dataset : string, required
        The path to the parameter file for the dataset to be analyzed.
    output_dir : string, optional
        If specified, all output will be put into this path instead of
        in the dataset directories.  Default: None.
    halos :  {"multiple", "single"}, optional
        For profiling more than one halo.  In this mode halos are read in
        from a list or identified with a halo finder.  In "single" mode,
        the one and only halo
        center is identified automatically as the location of the peak
        in the density field.
        Default: "multiple".
    halo_list_file : string, optional
        The name of a file containing the list of halos.  The HaloProfiler
        will  look for this file in the data directory.
        Default: "HopAnalysis.out".
    halo_list_format : {string, dict}
        The format of the halo list file.  "yt_hop" for the format
        given by yt's halo finders.  "enzo_hop" for the format written
        by enzo_hop. "p-groupfinder"  for P-Groupfinder.  This keyword
        can also be given in the form of a dictionary specifying the
        column in which various properties can be found.
        For example, {"id": 0, "center": [1, 2, 3], "mass": 4, "radius": 5}.
        Default: "yt_hop".
    halo_finder_function : function
        If halos is set to multiple and the file given by
        halo_list_file does not exit, the halo finding function
        specified here will be called.
        Default: HaloFinder (yt_hop).
    halo_finder_args : tuple
        Args given with call to halo finder function.  Default: None.
    halo_finder_kwargs : dict
        kwargs given with call to halo finder function. Default: None.
    recenter : {string, function}
        The exact location of the sphere center can significantly affect
        radial profiles.  The halo center loaded by the HaloProfiler will
        typically be the dark matter center of mass calculated by a halo
        finder.  However, this may not be the best location for centering
        profiles of baryon quantities.  For example, one may want to center
        on the maximum density.
        If recenter is given as a string, one of the existing recentering
        functions will be used:

            * Min_Dark_Matter_Density : location of minimum dark matter density
            * Max_Dark_Matter_Density : location of maximum dark matter density
            * CoM_Dark_Matter_Density : dark matter center of mass
            * Min_Gas_Density : location of minimum gas density
            * Max_Gas_Density : location of maximum gas density
            * CoM_Gas_Density : gas center of mass
            * Min_Total_Density : location of minimum total density
            * Max_Total_Density : location of maximum total density
            * CoM_Total_Density : total center of mass
            * Min_Temperature : location of minimum temperature
            * Max_Temperature : location of maximum temperature

        Alternately, a function can be supplied for custom recentering.
        The function should take only one argument, a sphere object.  Example
        function::
            
               def my_center_of_mass(data):
                   my_x, my_y, my_z = data.quantities['CenterOfMass']()
                   return (my_x, my_y, my_z)
            
        Default: None.
    halo_radius : float
        If no halo radii are provided in the halo list file, this
        parameter is used to specify the radius out to which radial
        profiles will be made.  This keyword is also
        used when halos is set to single.  Default: 0.1.
    radius_units : string
        The units of halo_radius.  Default: "1" (code units).
    n_profile_bins : int
        The number of bins in the radial profiles.  Default: 50.
    profile_output_dir : str
        The subdirectory, inside the data directory, in which radial profile
        output files will be created.  The directory will be created if it does
        not exist.  Default: "radial_profiles".
    projection_output_dir : str
        The subdirectory, inside the data directory, in which projection
        output files will be created.  The directory will be created if it does
        not exist.  Default: "projections".
    projection_width : float
        The width of halo projections.  Default: 8.0.
    projection_width_units : string
        The units of projection_width. Default: "mpc".
    project_at_level : {"max", int}
        The maximum refinement level to be included in projections.
        Default: "max" (maximum level within the dataset).
    velocity_center  : array_like
        The method in which the halo bulk velocity is calculated (used for
        calculation of radial and tangential velocities.  Valid options are:

            * ["bulk", "halo"] (Default): the velocity provided in
              the halo list
            * ["bulk", "sphere"]: the bulk velocity of the sphere
              centered on the halo center.
            * ["max", field]: the velocity of the cell that is the
              location of the maximum of the field specified.

    filter_quantities : array_like
        Quantities from the original halo list file to be written out in the
        filtered list file.  Default: ['id','center'].
    use_critical_density : bool
        If True, the definition of overdensity for virial quantities
        is calculated with respect to the critical density.
        If False, overdensity is with respect to mean matter density,
        which is lower by a factor of Omega_M.  Default: False.

    Examples
    --------

    >>> from yt.analysis_modules.halo_profiler.api import *
    >>> hp = HaloProfiler("RedshiftOutput0005/RD0005")

    """
    def __init__(self, dataset, output_dir=None,
                 halos='multiple', halo_list_file='HopAnalysis.out',
                 halo_list_format='yt_hop', halo_finder_function=parallelHF,
                 halo_finder_args=None,
                 halo_finder_kwargs=dict(threshold=160.0, safety=1.5,
                                         dm_only=False, resize=True,
                                         fancy_padding=True, rearrange=True),
                 halo_radius=None, radius_units='1', n_profile_bins=50,
                 recenter=None,
                 profile_output_dir='radial_profiles', projection_output_dir='projections',
                 projection_width=8.0, projection_width_units='mpc', project_at_level='max',
                 velocity_center=['bulk', 'halo'], filter_quantities=['id', 'center', 'r_max'],
                 use_critical_density=False):
        ParallelAnalysisInterface.__init__(self)

        self.dataset = dataset
        self.n_profile_bins = n_profile_bins
        self.projection_width = projection_width
        self.projection_width_units = projection_width_units
        self.project_at_level = project_at_level
        self.filter_quantities = filter_quantities
        if self.filter_quantities is None: self.filter_quantities = []
        self.use_critical_density = use_critical_density

        self.profile_fields = []
        self.projection_fields = []

        self._halo_filters = []
        self.all_halos = []
        self.filtered_halos = []

        # Set halo finder function and parameters, if needed.
        self.halo_finder_function = halo_finder_function
        self.halo_finder_args = halo_finder_args
        if self.halo_finder_args is None: self.halo_finder_args = ()
        self.halo_finder_kwargs = halo_finder_kwargs
        if self.halo_finder_kwargs is None: self.halo_finder_kwargs = {}

        # Set option to get halos from hop or single halo at density maximum.
        # multiple: get halos from hop
        # single: get single halo from density maximum
        self.halos = halos
        if not(self.halos is 'multiple' or self.halos is 'single'):
            mylog.error("Keyword, halos, must be either 'single' or 'multiple'.")
            return None

        # Set halo list format.
        # 'yt_hop': yt hop output.
        # 'enzo_hop': enzo_hop output.
        # dictionary: a dictionary containing fields and their corresponding columns.
        self.halo_list_file = halo_list_file
        if halo_list_format == 'yt_hop':
            self.halo_list_format = {'id':0, 'mass':1, 'np': 2,
                                     'center':[7, 8, 9], 'velocity':[10, 11, 12], 'r_max':13}
        elif halo_list_format == 'enzo_hop':
            self.halo_list_format = {'id':0, 'center':[4, 5, 6]}
        elif halo_list_format == 'p-groupfinder':
            self.halo_list_format = {'id':3, 'mass':5, 'center':[0, 1, 2], 'r200kpc':8}
        elif isinstance(halo_list_format, types.DictType):
            self.halo_list_format = halo_list_format
        else:
            mylog.error("Keyword, halo_list_format, must be 'yt_hop', 'enzo_hop', 'p-groupfinder', or a dictionary of custom settings.")
            return None

        # Option to recenter sphere someplace else.
        self.recenter = recenter

        # Flag for whether calculating halo bulk velocity is necessary.
        self._need_bulk_velocity = False

        # Check validity for VelocityCenter parameter which toggles how the
        # velocity is zeroed out for radial velocity profiles.
        self.velocity_center = velocity_center[:]
        if self.velocity_center[0] == 'bulk':
            if self.velocity_center[1] == 'halo' and \
                    self.halos is 'single':
                mylog.error("Parameter, VelocityCenter, must be set to 'bulk sphere' or 'max <field>' with halos flag set to 'single'.")
                return None
            if self.velocity_center[1] == 'halo' and \
                    self.halo_list_format is 'enzo_hop':
                mylog.error("Parameter, VelocityCenter, must be 'bulk sphere' for old style hop output files.")
                return None
            if not(self.velocity_center[1] == 'halo' or
                   self.velocity_center[1] == 'sphere'):
                mylog.error("Second value of VelocityCenter must be either 'halo' or 'sphere' if first value is 'bulk'.")
                return None
        elif self.velocity_center[0] == 'max':
            mylog.info('Using position of max %s for velocity center.' % self.velocity_center[1])
        else:
            mylog.error("First value of parameter, VelocityCenter, must be either 'bulk' or 'max'.")
            return None

        # Create dataset object.
        if isinstance(self.dataset, StaticOutput):
            self.pf = self.dataset
        else:
            self.pf = load(self.dataset)
        self.pf.h

        # Create output directories.
        self.output_dir = output_dir
        if output_dir is None:
            self.output_dir = self.pf.fullpath
        else:
            self.__check_directory(output_dir)
            self.output_dir = os.path.join(output_dir, self.pf.directory)
        self.__check_directory(self.output_dir)
        self.profile_output_dir = os.path.join(self.output_dir, profile_output_dir)
        self.projection_output_dir = os.path.join(self.output_dir, projection_output_dir)

        # Figure out what max radius to use for profiling.
        if halo_radius is not None:
            self.halo_radius = halo_radius / self.pf[radius_units]
        elif self.halos is 'single' or not 'r_max' in self.halo_list_format:
            self.halo_radius = 0.1
        else:
            self.halo_radius = None

        # Get halo(s).
        if self.halos is 'single':
            v, center = self.pf.h.find_max('Density')
            singleHalo = {}
            singleHalo['center'] = center
            singleHalo['r_max'] = self.halo_radius * self.pf.units['mpc']
            singleHalo['id'] = 0
            self.all_halos.append(singleHalo)
        elif self.halos is 'multiple':
            # Get hop data.
            self._load_halo_data()
            if len(self.all_halos) == 0:
                mylog.error("No halos loaded, there will be nothing to do.")
                return None
        else:
            mylog.error("Keyword, halos, must be either 'single' or 'multiple'.")
            return None

    def add_halo_filter(self, function, *args, **kwargs):
        r"""Filters can be added to create a refined list of halos based on
        their profiles or to avoid profiling halos altogether based on
        information given in the halo list file.

        It is often the case that one is looking to identify halos with a
        specific set of properties. This can be accomplished through the
        creation of filter functions. A filter function can take as many args
        and kwargs as you like, as long as the first argument is a profile
        object, or at least a dictionary which contains the profile arrays
        for each field. Filter functions must return a list of two things.
        The first is a True or False indicating whether the halo passed the
        filter. The second is a dictionary containing quantities calculated
        for that halo that will be written to a file if the halo passes the
        filter. A sample filter function based on virial quantities can be
        found in yt/analysis_modules/halo_profiler/halo_filters.py.

        Parameters
        ----------
        function : function
            The name of a halo filter function.
        args : values
            Arguments passed to the halo filter function.
        kwargs : values
            Arguments passed to the halo filter function.

        Examples
        -------
        >>> hp.add_halo_filter(HP.VirialFilter, must_be_virialized=True,
                overdensity_field='ActualOverdensity',
                virial_overdensity=200,
                virial_filters=[['TotalMassMsun','>=','1e14']],
                virial_quantities=['TotalMassMsun','RadiusMpc'])

        """

        self._halo_filters.append({'function':function, 'args':args, 'kwargs':kwargs})

    def add_profile(self, field, weight_field=None, accumulation=False):
        r"""Add a field for profiling.

        Once the halo profiler object has been instantiated,
        fields can be added for profiling using this function. This function
        may be called multiple times, once per field to be added.

        Parameters
        ----------
        field : string
            The name of the field.
        weight_field : {None, string}, optional
            The field that will be used to weight the field `field` when
            the radial binning is done. Default: None.
        accumulation : bool
            Whether or not the `field` values should be summed up with the
            radius of the profile.

        Examples
        >>> hp.add_profile('CellVolume', weight_field=None, accumulation=True)
        >>> hp.add_profile('TotalMassMsun', weight_field=None, accumulation=True)
        >>> hp.add_profile('Density', weight_field=None, accumulation=False)
        >>> hp.add_profile('Temperature', weight_field='CellMassMsun', accumulation=False)

        """

        # Check for any field that might need to have the bulk velocity set.
        if 'Velocity' in field or 'Mach' in field:
            self._need_bulk_velocity = True

        self.profile_fields.append({'field':field, 'weight_field':weight_field,
                                    'accumulation':accumulation})

    def add_projection(self, field, weight_field=None, cmap='algae'):
        r"""Make a projection of the specified field.

        For the given field, a projection will be produced that can be saved
        to HDF5 or image format. See `make_projections`.

        Parameters
        ----------
        field : string
            The name of the field.
        weight_field : string
            The field that will be used to weight the field `field` when
            the projection is done. Default: None.
        cmap : string
            The name of the matplotlib color map that will be used if an
            image is made from the projection. Default="algae".

        Examples
        --------
        >>> hp.add_projection('Density', weight_field=None)
        >>> hp.add_projection('Temperature', weight_field='Density')
        >>> hp.add_projection('Metallicity', weight_field='Density')

        """

        # Check for any field that might need to have the bulk velocity set.
        if 'Velocity' in field or 'Mach' in field:
            self._need_bulk_velocity = True

        self.projection_fields.append({'field':field, 'weight_field':weight_field,
                                       'cmap': cmap})

    @parallel_blocking_call
    def make_profiles(self, filename=None, prefilters=None, njobs=-1,
                      dynamic=False, profile_format='ascii'):
        r"""Make radial profiles for all halos in the list.

        After all the calls to `add_profile`, this will trigger the actual
        calculations and output the profiles to disk.

        Parameters
        ----------

        filename : str
            If set, a file will be written with all of the filtered halos
            and the quantities returned by the filter functions.
            Default: None.
        prefilters : array_like
            A single dataset can contain thousands or tens of thousands of
            halos. Significant time can be saved by not profiling halos
            that are certain to not pass any filter functions in place.
            Simple filters based on quantities provided in the initial
            halo list can be used to filter out unwanted halos using this
            parameter.
            Default: None.
        njobs : int
            The number of jobs over which to split the profiling.  Set
            to -1 so that each halo is done by a single processor.
            Default: -1.
        dynamic : bool
            If True, distribute halos using a task queue.  If False,
            distribute halos evenly over all jobs.
            Default: False.
        profile_format : str
            The file format for the radial profiles, 'ascii' or 'hdf5'.
            Default: 'ascii'.

        Examples
        --------

        >>> hp.make_profiles(filename="FilteredQuantities.out",
                 prefilters=["halo['mass'] > 1e13"])

        """

        extension_map = {'ascii': 'dat',
                         'hdf5': 'h5'}
        if not profile_format in extension_map:
            mylog.error("Invalid profile_format: %s.  Valid options are %s." %
                        (profile_format, ", ".join(extension_map.keys())))
            raise YTException(pf=self.pf)

        if len(self.all_halos) == 0:
            mylog.error("Halo list is empty, returning.")
            return None

        # Reset filtered halo list.
        self.filtered_halos = []

        # Check to see if the VirialFilter has been added to the filter list.
        # If a lower mass cutoff is being used, use it to make a pre-filter.
        if prefilters is None: prefilters = []
        virial_prefilter = None
        virial_prefilter_safety_factor = 0.5
        all_filter_functions = [hf['function'] for hf in self._halo_filters]
        virial_filter = VirialFilter in all_filter_functions
        if 'mass' in self.halo_list_format and VirialFilter in all_filter_functions:
            vFilter = self._halo_filters[all_filter_functions.index(VirialFilter)]
            if vFilter['kwargs'].has_key('virial_filters') and \
               vFilter['kwargs']['virial_filters'] is not None:
                all_vqFilters = [vqf[0] for vqf in vFilter['kwargs']['virial_filters']]
                if 'TotalMassMsun' in all_vqFilters:
                    mass_filter = vFilter['kwargs']['virial_filters'][all_vqFilters.index('TotalMassMsun')]
                    if '>' in mass_filter[1]:
                        virial_prefilter = "halo['mass'] %s %f * %s" % \
                            (mass_filter[1], virial_prefilter_safety_factor, mass_filter[2])
                        prefilters.append(virial_prefilter)
                    elif '<' in mass_filter[1]:
                        virial_prefilter = "halo['mass'] %s %f * %s" % \
                            (mass_filter[1], (1./virial_prefilter_safety_factor), mass_filter[2])
                        prefilters.append(virial_prefilter)

        # Add profile fields necessary for calculating virial quantities.
        if virial_filter: self._check_for_needed_profile_fields()

        # Create output directory.
        self.__check_directory(self.profile_output_dir)

        # Profile all halos.
        updated_halos = []
        for halo in parallel_objects(self.all_halos, njobs=njobs, dynamic=dynamic):
            # Apply prefilters to avoid profiling unwanted halos.
            filter_result = True
            haloQuantities = {}
            if prefilters is not None:
                for prefilter in prefilters:
                    if not eval(prefilter):
                        filter_result = False
                        break

            if filter_result and len(self.profile_fields) > 0:

                profile_filename = "%s/Halo_%04d_profile.%s" % \
                    (self.profile_output_dir, halo['id'], extension_map[profile_format])

                profiledHalo = self._get_halo_profile(halo, profile_filename,
                                                      virial_filter=virial_filter)

                if profiledHalo is None:
                    continue

                # Apply filter and keep track of the quantities that are returned.
                for hFilter in self._halo_filters:
                    filter_result, filterQuantities = hFilter['function'](profiledHalo, *hFilter['args'],
                                                                          **hFilter['kwargs'])

                    if not filter_result: break

                    if filterQuantities is not None:
                        haloQuantities.update(filterQuantities)

            if filter_result:
                for quantity in self.filter_quantities:
                    if halo.has_key(quantity): haloQuantities[quantity] = halo[quantity]

                only_on_root(self.filtered_halos.append, haloQuantities)

            # If we've gotten this far down, this halo is good and we want
            # to keep it. But we need to communicate the recentering changes
            # to all processors (the root one in particular) without having
            # one task clobber the other.
            only_on_root(updated_halos.append, halo)

        # And here is where we bring it all together.
        updated_halos = self.comm.par_combine_object(updated_halos,
                            datatype="list", op="cat")
        updated_halos.sort(key=lambda a:a['id'])
        self.all_halos = updated_halos

        self.filtered_halos = self.comm.par_combine_object(self.filtered_halos,
                            datatype="list", op="cat")
        self.filtered_halos.sort(key=lambda a:a['id'])

        if filename is not None:
            self._write_filtered_halo_list(os.path.join(self.output_dir, filename))

    def _get_halo_profile(self, halo, filename, virial_filter=True,
            force_write=False):
        r"""Profile a single halo and write profile data to a file.
        If file already exists, read profile data from file.
        Return a dictionary of id, center, and virial quantities if virial_filter is True.
        """

        # Read profile from file if it already exists.
        # If not, profile will be None.
        profile = self._read_profile(filename)

        # Make profile if necessary.
        newProfile = profile is None
        if newProfile:

            r_min = 2 * self.pf.h.get_smallest_dx() * self.pf['mpc']
            if (halo['r_max'] / r_min < PROFILE_RADIUS_THRESHOLD):
                mylog.debug("Skipping halo with r_max / r_min = %f." % (halo['r_max']/r_min))
                return None

            # get a sphere object to profile
            sphere = self._get_halo_sphere(halo)
            if sphere is None: return None

            try:
                profile = BinnedProfile1D(sphere, self.n_profile_bins, "RadiusMpc",
                                                r_min, halo['r_max'],
                                                log_space=True, lazy_reader=True,
                                                end_collect=True)
            except EmptyProfileData:
                mylog.error("Caught EmptyProfileData exception, returning None for this halo.")
                return None
            # Figure out which fields to add simultaneously
            field_groupings = defaultdict(lambda: defaultdict(list))
            for hp in self.profile_fields:
                field_groupings[hp['weight_field']][hp['accumulation']].append(hp['field'])
            for weight_field in field_groupings:
                for accum, fields in field_groupings[weight_field].items():
                    profile.add_fields(fields, weight=weight_field,
                                       accumulation=accum)

        if virial_filter:
            self._add_actual_overdensity(profile)

        if newProfile:
            mylog.info("Writing halo %d" % halo['id'])
            if os.path.exists(filename): os.remove(filename)
            if filename.endswith('.h5'):
                profile.write_out_h5(filename)
            else:
                profile.write_out(filename, format='%0.6e')
            # profile will have N+1 bins so remove the last one
            for field in profile.keys():
                profile[field] = profile[field][:-1]
        elif force_write:
            mylog.info("Re-writing halo %d" % halo['id'])
            self._write_profile(profile, filename, format='%0.6e')

        if newProfile:
            # Temporary solution to memory leak.
            for g in self.pf.h.grids:
                g.clear_data()
            sphere.clear_data()
            del sphere
            # Currently, this seems to be the only way to prevent large 
            # halo profiling runs from running out of ram.
            # It would be good to track down the real cause at some point.
            gc.collect()

        return profile

    def _get_halo_sphere(self, halo):
        """
        Returns a sphere object for a given halo, performs the recentering,
        and calculates bulk velocities.
        """

        sphere = self.pf.h.sphere(halo['center'], halo['r_max']/self.pf.units['mpc'])
        if len(sphere._grids) == 0: return None
        new_sphere = False

        if self.recenter:
            old = halo['center']
            if self.recenter in centering_registry:
                new_x, new_y, new_z = \
                    centering_registry[self.recenter](sphere)
            else:
                # user supplied function
                new_x, new_y, new_z = self.recenter(sphere)
            if new_x < self.pf.domain_left_edge[0] or \
                    new_y < self.pf.domain_left_edge[1] or \
                    new_z < self.pf.domain_left_edge[2]:
                mylog.info("Recentering rejected, skipping halo %d" % \
                    halo['id'])
                return None
            halo['center'] = [new_x, new_y, new_z]
            d = self.pf['kpc'] * periodic_dist(old, halo['center'],
                self.pf.domain_right_edge - self.pf.domain_left_edge)
            mylog.info("Recentered halo %d %1.3e kpc away." % (halo['id'], d))
            # Expand the halo to account for recentering.
            halo['r_max'] += d / 1000. # d is in kpc -> want mpc
            new_sphere = True

        if new_sphere:
            # Temporary solution to memory leak.
            for g in self.pf.h.grids:
                g.clear_data()
            sphere.clear_data()
            del sphere
            sphere = self.pf.h.sphere(halo['center'], halo['r_max']/self.pf.units['mpc'])

        if self._need_bulk_velocity:
            # Set bulk velocity to zero out radial velocity profiles.
            if self.velocity_center[0] == 'bulk':
                if self.velocity_center[1] == 'halo':
                    sphere.set_field_parameter('bulk_velocity', halo['velocity'])
                elif self.velocity_center[1] == 'sphere':
                    mylog.info('Calculating sphere bulk velocity.')
                    sphere.set_field_parameter('bulk_velocity',
                                               sphere.quantities['BulkVelocity'](lazy_reader=True,
                                                                                 preload=False))
                else:
                    mylog.error("Invalid parameter: velocity_center.")
                    return None
            elif self.velocity_center[0] == 'max':
                mylog.info('Setting bulk velocity with value at max %s.' % self.velocity_center[1])
                max_val, maxi, mx, my, mz, mg = sphere.quantities['MaxLocation'](self.velocity_center[1],
                                                                                 lazy_reader=True)
                max_grid = self.pf.h.grids[mg]
                max_cell = np.unravel_index(maxi, max_grid.ActiveDimensions)
                sphere.set_field_parameter('bulk_velocity', [max_grid['x-velocity'][max_cell],
                                                             max_grid['y-velocity'][max_cell],
                                                             max_grid['z-velocity'][max_cell]])
            mylog.info('Bulk velocity set.')

        return sphere

    @parallel_blocking_call
    def make_projections(self, axes=[0, 1, 2], halo_list='filtered',
                         save_images=False, save_cube=True, njobs=-1,
                         dynamic=False):
        r"""Make projections of all halos using specified fields.

        After adding fields using `add_projection`, this starts the actual
        calculations and saves the output to disk.

        Parameters
        ----------

        axes : array_like
            A list of the axes to project along, using the usual 0,1,2
            convention. Default=[0,1,2]
        halo_list : {'filtered', 'all'}
            Which set of halos to make profiles of, either ones passed by the
            halo filters (if enabled/added), or all halos.
            Default='filtered'.
        save_images : bool
            Whether or not to save images of the projections. Default=False.
        save_cube : bool
            Whether or not to save the HDF5 files of the halo projections.
            Default=True.
        njobs : int
            The number of jobs over which to split the projections.  Set
            to -1 so that each halo is done by a single processor.  Halo 
            projections do not currently work in parallel, so this must 
            be set to -1.
            Default: -1.
        dynamic : bool
            If True, distribute halos using a task queue.  If False,
            distribute halos evenly over all jobs.
            Default: False.

        Examples
        --------

        >>> hp.make_projections(axes=[0, 1, 2], save_cube=True,
            save_images=True, halo_list="filtered")

        """

        # Halo projections cannot run in parallel because they are done by 
        # giving a data source to the projection object.
        if njobs > 0:
            mylog.warn("Halo projections cannot use more than one processor per halo, setting njobs to -1.")
            njobs = -1
        
        # Get list of halos for projecting.
        if halo_list == 'filtered':
            halo_projection_list = self.filtered_halos
        elif halo_list == 'all':
            halo_projection_list = self.all_halos
        elif isinstance(halo_list, types.StringType):
            halo_projection_list = self._read_halo_list(halo_list)
        elif isinstance(halo_list, types.ListType):
            halo_projection_list = halo_list
        else:
            mylog.error("Keyword, halo_list', must be 'filtered', 'all', a filename, or an actual list.")
            return

        if len(halo_projection_list) == 0:
            mylog.error("Halo list for projections is empty.")
            return

        # Set resolution for fixed resolution output.
        if self.project_at_level == 'max':
            proj_level = self.pf.h.max_level
        else:
            proj_level = int(self.project_at_level)
        proj_dx = self.pf.units[self.projection_width_units] / \
            self.pf.parameters['TopGridDimensions'][0] / \
            (self.pf.parameters['RefineBy']**proj_level)
        projectionResolution = int(self.projection_width / proj_dx)

        # Create output directory.
        self.__check_directory(self.projection_output_dir)

        center = [0.5 * (self.pf.parameters['DomainLeftEdge'][w] +
                         self.pf.parameters['DomainRightEdge'][w])
                  for w in range(self.pf.parameters['TopGridRank'])]

        for halo in parallel_objects(halo_projection_list, njobs=njobs, dynamic=dynamic):
            if halo is None:
                continue
            # Check if region will overlap domain edge.
            # Using non-periodic regions is faster than using periodic ones.
            leftEdge = [(halo['center'][w] -
                         0.5 * self.projection_width/self.pf.units[self.projection_width_units])
                        for w in range(len(halo['center']))]
            rightEdge = [(halo['center'][w] +
                          0.5 * self.projection_width/self.pf.units[self.projection_width_units])
                         for w in range(len(halo['center']))]

            mylog.info("Projecting halo %04d in region: [%f, %f, %f] to [%f, %f, %f]." %
                       (halo['id'], leftEdge[0], leftEdge[1], leftEdge[2],
                        rightEdge[0], rightEdge[1], rightEdge[2]))

            need_per = False
            for w in range(len(halo['center'])):
                if ((leftEdge[w] < self.pf.parameters['DomainLeftEdge'][w]) or
                    (rightEdge[w] > self.pf.parameters['DomainRightEdge'][w])):
                    need_per = True
                    break

            if need_per:
                region = self.pf.h.periodic_region(halo['center'], leftEdge, rightEdge)
            else:
                region = self.pf.h.region(halo['center'], leftEdge, rightEdge)

            # Make projections.
            if not isinstance(axes, types.ListType): axes = list([axes])
            for w in axes:
                projections = []
                # YT projections do not follow the right-hand rule.
                coords = range(3)
                del coords[w]
                x_axis = coords[0]
                y_axis = coords[1]

                for hp in self.projection_fields:
                    projections.append(self.pf.h.proj(w, hp['field'],
                                                      weight_field=hp['weight_field'],
                                                      source=region, center=halo['center'],
                                                      serialize=False))

                # Set x and y limits, shift image if it overlaps domain boundary.
                if need_per:
                    pw = self.projection_width/self.pf.units[self.projection_width_units]
                    _shift_projections(self.pf, projections, halo['center'], center, w)
                    # Projection has now been shifted to center of box.
                    proj_left = [center[x_axis]-0.5*pw, center[y_axis]-0.5*pw]
                    proj_right = [center[x_axis]+0.5*pw, center[y_axis]+0.5*pw]
                else:
                    proj_left = [leftEdge[x_axis], leftEdge[y_axis]]
                    proj_right = [rightEdge[x_axis], rightEdge[y_axis]]

                # Save projection data to hdf5 file.
                if save_cube or save_images:
                    axis_labels = ['x', 'y', 'z']

                    if save_cube:
                        dataFilename = "%s/Halo_%04d_%s_data.h5" % \
                            (self.projection_output_dir, halo['id'], axis_labels[w])
                        mylog.info("Saving projection data to %s." % dataFilename)
                        output = h5py.File(dataFilename, "a")

                    # Create fixed resolution buffer for each projection and write them out.
                    for e, hp in enumerate(self.projection_fields):
                        frb = FixedResolutionBuffer(projections[e], (proj_left[0], proj_right[0],
                                                                     proj_left[1], proj_right[1]),
                                                    (projectionResolution, projectionResolution),
                                                    antialias=False)
                        dataset_name = "%s_%s" % (hp['field'], hp['weight_field'])
                        if save_cube:
                            if dataset_name in output: del output[dataset_name]
                            output.create_dataset(dataset_name, data=frb[hp['field']])
                        if save_images:
                            filename = "%s/Halo_%04d_%s_%s.png" % \
                              (self.projection_output_dir, halo['id'],
                               dataset_name, axis_labels[w])
                            if (frb[hp['field']] != 0).any():
                                write_image(np.log10(frb[hp['field']]), filename, cmap_name=hp['cmap'])
                            else:
                                mylog.info('Projection of %s for halo %d is all zeros, skipping image.' %
                                            (hp['field'], halo['id']))
                    if save_cube: output.close()

            del region

    @parallel_blocking_call
    def analyze_halo_spheres(self, analysis_function, halo_list='filtered',
                             analysis_output_dir=None, njobs=-1, dynamic=False):
        r"""Perform custom analysis on all halos.

        This will loop through all halo on the HaloProfiler's list,
        creating a sphere object for each halo and passing that sphere
        to the provided analysis function.

        Parameters
        ----------

        analysis_function : function
            A function taking two arguments, the halo dictionary, and a
            sphere object.
            Example function to calculate total mass of halo::

                def my_analysis(halo, sphere):
                    total_mass = sphere.quantities['TotalMass']()
                    print total_mass
            
        halo_list : {'filtered', 'all'}
            Which set of halos to make profiles of, either ones passed by the
            halo filters (if enabled/added), or all halos.
            Default='filtered'.
        analysis_output_dir : string, optional
            If specified, this directory will be created within the dataset to
            contain any output from the analysis function.  Default: None.
        njobs : int
            The number of jobs over which to split the analysis.  Set
            to -1 so that each halo is done by a single processor.
            Default: -1.
        dynamic : bool
            If True, distribute halos using a task queue.  If False,
            distribute halos evenly over all jobs.
            Default: False.

        Examples
        --------

        >>> hp.analyze_halo_spheres(my_analysis, halo_list="filtered",
                                    analysis_output_dir='special_analysis')

        """

        # Get list of halos for projecting.
        if halo_list == 'filtered':
            halo_analysis_list = self.filtered_halos
        elif halo_list == 'all':
            halo_analysis_list = self.all_halos
        elif isinstance(halo_list, types.StringType):
            halo_analysis_list = self._read_halo_list(halo_list)
        elif isinstance(halo_list, types.ListType):
            halo_analysis_list = halo_list
        else:
            mylog.error("Keyword, halo_list', must be 'filtered', 'all', a filename, or an actual list.")
            return

        if len(halo_analysis_list) == 0:
            mylog.error("Halo list for analysis is empty.")
            return

        # Create output directory.
        if analysis_output_dir is not None:
            my_output_dir = os.path.join(self.output_dir, analysis_output_dir)
            self.__check_directory(my_output_dir)

        for halo in parallel_objects(halo_analysis_list, njobs=njobs, dynamic=dynamic):
            if halo is None: continue

            # Get a sphere object to analze.
            sphere = self._get_halo_sphere(halo)
            if sphere is None: continue

            # Call the given analysis function.
            analysis_function(halo, sphere)

    def _add_actual_overdensity(self, profile):
        "Calculate overdensity from TotalMassMsun and CellVolume fields."

        if 'ActualOverdensity' in profile.keys():
            return

        rho_crit_now = 1.8788e-29 * self.pf.hubble_constant**2 # g cm^-3
        Msun2g = 1.989e33
        rho_crit = rho_crit_now * ((1.0 + self.pf.current_redshift)**3.0)
        if not self.use_critical_density: rho_crit *= self.pf.omega_matter

        profile['ActualOverdensity'] = (Msun2g * profile['TotalMassMsun']) / \
            profile['CellVolume'] / rho_crit

    def _check_for_needed_profile_fields(self):
        "Make sure CellVolume and TotalMass fields are added so virial quantities can be calculated."
        all_profile_fields = [hp['field'] for hp in self.profile_fields]
        if not 'CellVolume' in all_profile_fields:
            mylog.info("Adding CellVolume field to so virial quantities can be calculated")
            self.add_profile('CellVolume', weight_field=None, accumulation=True)
        if not 'TotalMassMsun' in all_profile_fields:
            mylog.info("Adding TotalMassMsun field to so virial quantities can be calculated")
            self.add_profile('TotalMassMsun', weight_field=None, accumulation=True)

    def _load_halo_data(self, filename=None):
        "Read hop output file or run hop if it doesn't exist."

        # Don't run if hop data already loaded.
        if self.all_halos:
            return

        if filename is None:
            filename = self.halo_list_file

        hop_file = os.path.join(self.output_dir, filename)

        if not(os.path.exists(hop_file)):
            mylog.info("Halo finder file not found, running halo finder to get halos.")
            self._run_hop(hop_file)

        self.all_halos = self._read_halo_list(hop_file)

    def _read_halo_list(self, listFile):
        """
        Read halo list from aue file.
        Allow for columnar data in varying formats.
        """

        def __isE(arg):
            parts = arg.lower().split('e')
            if len(parts) != 2: return False
            return not (True in [q.isalpha() for q in ''.join(parts)])

        def __get_num(arg):
            if __isE(arg):
                return float(arg)
            if arg != arg.swapcase():
                return arg
            return float(arg)

        mylog.info("Reading halo information from %s." % listFile)
        haloList = []
        listLines = file(listFile)

        fields = self.halo_list_format.keys()
        getID = not 'id' in fields
        has_rmax = 'r_max' in fields
        has_r200kpc = 'r200kpc' in fields

        for line in listLines:
            line = line.strip()
            if len(line) > 0 and not line.startswith('#') and not line[0].isalpha():
                halo = {}
                onLine = line.split()
                for field in fields:
                    if isinstance(self.halo_list_format[field], types.ListType):
                        halo[field] = [__get_num(onLine[q]) for q in self.halo_list_format[field]]
                    else:
                        halo[field] = __get_num(onLine[self.halo_list_format[field]])
                if getID: halo['id'] = len(haloList)
                if self.halo_radius is not None:
                    halo['r_max'] = self.halo_radius * self.pf.units['mpc']
                elif has_rmax:
                    halo['r_max'] *= self.pf.units['mpc']
                elif has_r200kpc:
                    # If P-Groupfinder used, r_200 [kpc] is calculated.
                    # set r_max as 50% past r_200.
                    halo['r_max'] = 1.5 * halo['r200kpc'] / 1000.
                else:
                    mylog.error("HaloProfiler has no way to get halo radius.")
                    return None
                haloList.append(halo)

        mylog.info("Loaded %d halos." % (len(haloList)))
        return haloList

    def _read_profile(self, profileFile):
        "Read radial profile from file.  Return None if it doesn't have all the fields requested."

        # Check to see if file exists.
        if not os.path.exists(profileFile):
            return None

        if profileFile.endswith('.h5'):
            return self._read_profile_hdf5(profileFile)
        else:
            return self._read_profile_ascii(profileFile)

    def _read_profile_ascii(self, profileFile):
        "Read radial profile from file.  Return None if it doesn't have all the fields requested."

        f = open(profileFile, 'r')
        lines = f.readlines()
        f.close()

        if not lines:
            return None

        # Get fields from header.
        header = lines.pop(0)
        header = header.strip()
        fields = header.split()
        # First string is '#'.
        fields.pop(0)

        profile = {}
        profile_obj = FakeProfile(self.pf)
        for field in fields:
            profile[field] = []

        # Check if all fields needed are present.
        all_profile_fields = [hp['field'] for hp in self.profile_fields]
        for field in all_profile_fields:
            if not field in profile:
                return None

        # Fill profile fields, skip bad values.
        for line in lines:
            line = line.strip()
            onLine = line.split()
            lineOK = True
            for value in onLine:
                if value.isalpha():
                    lineOK = False
                    break
            if lineOK:
                for q, field in enumerate(fields):
                    profile[field].append(float(onLine[q]))

        for field in fields:
            profile[field] = np.array(profile[field])

        profile_obj._data = profile

        if len(profile[fields[0]]) > 1:
            return profile_obj
        else:
            return None

    def _read_profile_hdf5(self, profileFile):
        "Read radial profile from file.  Return None if it doesn't have all the fields requested."

        profile = {}
        try:
            in_file = h5py.File(profileFile, 'r')
        except IOError:
            return None
        if not 'RadiusMpc-1d' in in_file:
            return None
        my_group = in_file['RadiusMpc-1d']
        if not 'x-axis-RadiusMpc' in my_group.attrs:
            return None
        profile['RadiusMpc'] = my_group.attrs['x-axis-RadiusMpc']
        fields = my_group.keys()

        # Check if all fields needed are present.
        all_profile_fields = [hp['field'] for hp in self.profile_fields]
        for field in all_profile_fields:
            if not field in fields:
                in_file.close()
                return None

        for field in fields:
            profile[field] = my_group[field][:]
        in_file.close()

        profile_obj = FakeProfile(self.pf)
        profile_obj._data = profile
        return profile_obj

    @parallel_blocking_call
    def _run_hop(self, hop_file):
        "Run hop to get halos."

        hop_results = self.halo_finder_function(self.pf, *self.halo_finder_args,
                                                **self.halo_finder_kwargs)
        hop_results.write_out(hop_file)

        del hop_results
        self.pf.h.clear_all_data()

    @parallel_root_only
    def _write_filtered_halo_list(self, filename, format="%s"):
        """
        Write out list of filtered halos along with any quantities
        picked up during the filtering process.
        """

        if filename.endswith('.h5'):
            self._write_filtered_halo_list_h5(filename)
        else:
            self._write_filtered_halo_list_ascii(filename, format=format)

    def _write_filtered_halo_list_ascii(self, filename, format="%s"):
        """
        Write out list of filtered halos along with any quantities
        picked up during the filtering process.
        """

        if len(self.filtered_halos) == 0:
            mylog.error("No halos in filtered list.")
            return

        mylog.info("Writing filtered halo list to %s." % filename)
        out_file = open(filename, "w")
        fields = [field for field in sorted(self.filtered_halos[0])]
        halo_fields = []
        for halo_field in self.filter_quantities:
            if halo_field in fields:
                fields.remove(halo_field)
                halo_fields.append(halo_field)
        # Make it so number of fields in header is same as number of data columns.
        header_fields = []
        for halo_field in halo_fields:
            if isinstance(self.filtered_halos[0][halo_field], types.ListType):
                header_fields.extend(["%s[%d]" % (halo_field, q)
                                      for q in range(len(self.filtered_halos[0][halo_field]))])
            else:
                header_fields.append(halo_field)
        out_file.write("# ")
        out_file.write("\t".join(header_fields + fields + ["\n"]))

        for halo in self.filtered_halos:
            for halo_field in halo_fields:
                if isinstance(halo[halo_field], types.ListType):
                    field_data = np.array(halo[halo_field])
                    field_data.tofile(out_file, sep="\t", format=format)
                else:
                    if halo_field == 'id':
                        out_file.write("%04d" % halo[halo_field])
                    else:
                        out_file.write("%s" % halo[halo_field])
                out_file.write("\t")
            field_data = np.array([halo[field] for field in fields])
            field_data.tofile(out_file, sep="\t", format=format)
            out_file.write("\n")
        out_file.close()

    def _write_filtered_halo_list_h5(self, filename):
        """
        Write out list of filtered halos along with any quantities
        picked up during the filtering process.
        """

        if len(self.filtered_halos) == 0:
            mylog.error("No halos in filtered list.")
            return

        mylog.info("Writing filtered halo list to %s." % filename)
        out_file = h5py.File(filename, "w")
        fields = [field for field in sorted(self.filtered_halos[0])]
        halo_fields = []
        for halo_field in self.filter_quantities:
            if halo_field in fields:
                fields.remove(halo_field)
                halo_fields.append(halo_field)

        for halo_field in halo_fields + fields:
            value_list = []
            for halo in self.filtered_halos:
                value_list.append(halo[halo_field])
            value_list = np.array(value_list)
            out_file.create_dataset(halo_field, data=value_list)
        out_file.close()

    def _write_profile(self, profile, filename, format="%0.16e"):
        fid = open(filename, "w")
        fields = [field for field in sorted(profile.keys()) if field != "UsedBins"]
        fid.write("\t".join(["#"] + fields + ["\n"]))
        field_data = np.array([profile[field] for field in fields])
        for line in range(field_data.shape[1]):
            field_data[:, line].tofile(fid, sep="\t", format=format)
            fid.write("\n")
        fid.close()

    @parallel_root_only
    def __check_directory(self, my_output_dir):
        if (os.path.exists(my_output_dir)):
            if not(os.path.isdir(my_output_dir)):
                mylog.error("Output directory exists, but is not a directory: %s." % my_output_dir)
                raise IOError(my_output_dir)
        else:
            os.mkdir(my_output_dir)

def _shift_projections(pf, projections, oldCenter, newCenter, axis):
    """
    Shift projection data around.
    This is necessary when projecting a preiodic region.
    """
    offset = [newCenter[q]-oldCenter[q] for q in range(len(oldCenter))]
    width = [pf.parameters['DomainRightEdge'][q]-pf.parameters['DomainLeftEdge'][q] \
                 for q in range(len(oldCenter))]

    del offset[axis]
    del width[axis]

    for plot in projections:
        # Get name of data field.
        other_fields = {'px':True, 'py':True, 'pdx':True, 'pdy':True, 'weight_field':True}
        for pfield in plot.field_data.keys():
            if not(other_fields.has_key(pfield)):
                field = pfield
                break

        # Shift x and y positions.
        plot['px'] += offset[0]
        plot['py'] += offset[1]

        # Wrap off-edge cells back around to other side (periodic boundary conditions).
        plot['px'][plot['px'] < 0] += width[0]
        plot['py'][plot['py'] < 0] += width[1]
        plot['px'][plot['px'] > width[0]] -= width[0]
        plot['py'][plot['py'] > width[1]] -= width[1]

        # After shifting, some cells have fractional coverage on both sides of the box.
        # Find those cells and make copies to be placed on the other side.

        # Cells hanging off the right edge.
        add_x_px = plot['px'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_px -= width[0]
        add_x_py = plot['py'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_pdx = plot['pdx'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_pdy = plot['pdy'][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_field = plot[field][plot['px'] + 0.5 * plot['pdx'] > width[0]]
        add_x_weight_field = plot['weight_field'][plot['px'] + 0.5 * plot['pdx'] > width[0]]

        # Cells hanging off the left edge.
        add2_x_px = plot['px'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_px += width[0]
        add2_x_py = plot['py'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_pdx = plot['pdx'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_pdy = plot['pdy'][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_field = plot[field][plot['px'] - 0.5 * plot['pdx'] < 0]
        add2_x_weight_field = plot['weight_field'][plot['px'] - 0.5 * plot['pdx'] < 0]

        # Cells hanging off the top edge.
        add_y_px = plot['px'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_py = plot['py'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_py -= width[1]
        add_y_pdx = plot['pdx'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_pdy = plot['pdy'][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_field = plot[field][plot['py'] + 0.5 * plot['pdy'] > width[1]]
        add_y_weight_field = plot['weight_field'][plot['py'] + 0.5 * plot['pdy'] > width[1]]

        # Cells hanging off the bottom edge.
        add2_y_px = plot['px'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_py = plot['py'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_py += width[1]
        add2_y_pdx = plot['pdx'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_pdy = plot['pdy'][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_field = plot[field][plot['py'] - 0.5 * plot['pdy'] < 0]
        add2_y_weight_field = plot['weight_field'][plot['py'] - 0.5 * plot['pdy'] < 0]

        # Add the hanging cells back to the projection data.
        plot.field_data['px'] = np.concatenate([plot['px'], add_x_px, add_y_px,
                                                add2_x_px, add2_y_px])
        plot.field_data['py'] = np.concatenate([plot['py'], add_x_py, add_y_py,
                                                add2_x_py, add2_y_py])
        plot.field_data['pdx'] = np.concatenate([plot['pdx'], add_x_pdx, add_y_pdx,
                                                 add2_x_pdx, add2_y_pdx])
        plot.field_data['pdy'] = np.concatenate([plot['pdy'], add_x_pdy, add_y_pdy,
                                                 add2_x_pdy, add2_y_pdy])
        plot.field_data[field] = np.concatenate([plot[field], add_x_field, add_y_field,
                                                 add2_x_field, add2_y_field])
        plot.field_data['weight_field'] = np.concatenate([plot['weight_field'],
                                                          add_x_weight_field, add_y_weight_field,
                                                          add2_x_weight_field, add2_y_weight_field])

        # Delete original copies of hanging cells.
        del add_x_px, add_y_px, add2_x_px, add2_y_px
        del add_x_py, add_y_py, add2_x_py, add2_y_py
        del add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx
        del add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy
        del add_x_field, add_y_field, add2_x_field, add2_y_field
        del add_x_weight_field, add_y_weight_field, add2_x_weight_field, add2_y_weight_field

class FakeProfile(ParallelAnalysisInterface):
    """
    This is used to mimic a profile object when reading profile data from disk.
    """
    def __init__(self, pf):
        ParallelAnalysisInterface.__init__(self)
        self.pf = pf
        self._data = {}

    def __getitem__(self, key):
        return self._data[key]

    def keys(self):
        return self._data.keys()

standard_fields = [
    ("Density", "CellMassMsun", False),
    ("Temperature", "CellMassMsun", False),
    ("VelocityMagnitude", "CellMassMsun", False),
    ("Ones", None, False),
    ("Entropy", "CellMassMsun", False),
    ("RadialVelocity", "CellMassMsun", False),
    ("SpecificAngularMomentumX", "CellMassMsun", False),
    ("SpecificAngularMomentumY", "CellMassMsun", False),
    ("SpecificAngularMomentumZ", "CellMassMsun", False),
    ("CoolingTime", "CellMassMsun", False),
    ("DynamicalTime", "CellMassMsun", False),
    ("CellMassMsun", None, True),
    ("TotalMassMsun", None, True),
    ("Dark_Matter_Density", "CellMassMsun", False),
    #("ParticleSpecificAngularMomentumX", "ParticleMassMsun"),
    #("ParticleSpecificAngularMomentumY", "ParticleMassMsun"),
    #("ParticleSpecificAngularMomentumZ", "ParticleMassMsun"),
    ("OverDensity", "CellMassMsun", False),
    #("ParticleMassMsun", None),
    ("StarParticleDensity", "StarParticleMassMsun", False), # How do we weight this?
    #("StarParticleMassMsun", None),
    ("StarParticleDensity", "StarParticleMassMsun", False), # How do we weight this?
]

standard_fields += [("%s_Fraction" % (s), "CellMassMsun", False)
    for s in ["HI","HII","HeI","HeII","HeIII","H2I","H2II",
    "HM","Electron", "DI","DII","HDI","Metal"]
]



"""
Generalized Enzo output objects, both static and time-series.

Presumably at some point EnzoRun will be absorbed into here.
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk, J. S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import string, re, gc, time, os, os.path, weakref

from yt.funcs import *

from yt.config import ytcfg
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_root_only
from yt.utilities.parameter_file_storage import \
    ParameterFileStore, \
    NoParameterShelf, \
    output_type_registry
from yt.data_objects.field_info_container import \
    FieldInfoContainer, NullFunc
from yt.utilities.minimal_representation import \
    MinimalStaticOutput

# We want to support the movie format in the future.
# When such a thing comes to pass, I'll move all the stuff that is contant up
# to here, and then have it instantiate EnzoStaticOutputs as appropriate.

_cached_pfs = weakref.WeakValueDictionary()
_pf_store = ParameterFileStore()

class StaticOutput(object):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            output_type_registry[name] = cls
            mylog.debug("Registering: %s as %s", name, cls)

    def __new__(cls, filename=None, *args, **kwargs):
        if not isinstance(filename, types.StringTypes):
            obj = object.__new__(cls)
            obj.__init__(filename, *args, **kwargs)
            return obj
        apath = os.path.abspath(filename)
        if not os.path.exists(apath): raise IOError(filename)
        if apath not in _cached_pfs:
            obj = object.__new__(cls)
            _cached_pfs[apath] = obj
        return _cached_pfs[apath]

    def __init__(self, filename, data_style=None, file_style=None):
        """
        Base class for generating new output types.  Principally consists of
        a *filename* and a *data_style* which will be passed on to children.
        """
        self.data_style = data_style
        self.file_style = file_style
        self.conversion_factors = {}
        self.parameters = {}

        # path stuff
        self.parameter_filename = str(filename)
        self.basename = os.path.basename(filename)
        self.directory = os.path.expanduser(os.path.dirname(filename))
        self.fullpath = os.path.abspath(self.directory)
        self.backup_filename = self.parameter_filename + '_backup.gdf'
        self.read_from_backup = False
        if os.path.exists(self.backup_filename):
            self.read_from_backup = True
        if len(self.directory) == 0:
            self.directory = "."

        # to get the timing right, do this before the heavy lifting
        self._instantiated = time.time()

        self.min_level = 0

        self._parse_parameter_file()
        self._set_units()
        self._set_derived_attrs()

        # Because we need an instantiated class to check the pf's existence in
        # the cache, we move that check to here from __new__.  This avoids
        # double-instantiation.
        try:
            _pf_store.check_pf(self)
        except NoParameterShelf:
            pass
        self.print_key_parameters()

        self.create_field_info()

    def _set_derived_attrs(self):
        self.domain_center = 0.5 * (self.domain_right_edge + self.domain_left_edge)
        self.domain_width = self.domain_right_edge - self.domain_left_edge

    def __reduce__(self):
        args = (self._hash(),)
        return (_reconstruct_pf, args)

    def __repr__(self):
        return self.basename

    def _hash(self):
        s = "%s;%s;%s" % (self.basename,
            self.current_time, self.unique_identifier)
        try:
            import hashlib
            return hashlib.md5(s).hexdigest()
        except ImportError:
            return s.replace(";", "*")

    @property
    def _mrep(self):
        return MinimalStaticOutput(self)

    def hub_upload(self):
        self._mrep.upload()

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        return False

    def __getitem__(self, key):
        """ Returns units, parameters, or conversion_factors in that order. """
        for d in [self.units, self.time_units, self.parameters, \
                  self.conversion_factors]:
            if key in d: return d[key]
        raise KeyError(key)

    def keys(self):
        """
        Returns a list of possible keys, from units, parameters and
        conversion_factors.

        """
        return self.units.keys() \
             + self.time_units.keys() \
             + self.parameters.keys() \
             + self.conversion_factors.keys()

    def __iter__(self):
        for ll in [self.units, self.time_units,
                   self.parameters, self.conversion_factors]:
            for i in ll.keys(): yield i

    def get_smallest_appropriate_unit(self, v):
        max_nu = 1e30
        good_u = None
        for unit in ['mpc', 'kpc', 'pc', 'au', 'rsun', 'km', 'cm']:
            vv = v*self[unit]
            if vv < max_nu and vv > 1.0:
                good_u = unit
                max_nu = v*self[unit]
        if good_u is None : good_u = 'cm'
        return good_u

    def has_key(self, key):
        """
        Checks units, parameters, and conversion factors. Returns a boolean.

        """
        return key in self.units or \
               key in self.time_units or \
               key in self.parameters or \
               key in self.conversion_factors

    _instantiated_hierarchy = None
    @property
    def hierarchy(self):
        if self._instantiated_hierarchy is None:
            if self._hierarchy_class == None:
                raise RuntimeError("You should not instantiate StaticOutput.")
            self._instantiated_hierarchy = self._hierarchy_class(
                self, data_style=self.data_style)
        return self._instantiated_hierarchy
    h = hierarchy  # alias

    @parallel_root_only
    def print_key_parameters(self):
        for a in ["current_time", "domain_dimensions", "domain_left_edge",
                  "domain_right_edge", "cosmological_simulation"]:
            if not hasattr(self, a):
                mylog.error("Missing %s in parameter file definition!", a)
                continue
            v = getattr(self, a)
            mylog.info("Parameters: %-25s = %s", a, v)
        if hasattr(self, "cosmological_simulation") and \
           getattr(self, "cosmological_simulation"):
            for a in ["current_redshift", "omega_lambda", "omega_matter",
                      "hubble_constant"]:
                if not hasattr(self, a):
                    mylog.error("Missing %s in parameter file definition!", a)
                    continue
                v = getattr(self, a)
                mylog.info("Parameters: %-25s = %s", a, v)

    def create_field_info(self):
        if getattr(self, "field_info", None) is None:
            # The setting up of fields occurs in the hierarchy, which is only
            # instantiated once.  So we have to double check to make sure that,
            # in the event of double-loads of a parameter file, we do not blow
            # away the exising field_info.
            self.field_info = FieldInfoContainer.create_with_fallback(
                                self._fieldinfo_fallback)

        

def _reconstruct_pf(*args, **kwargs):
    pfs = ParameterFileStore()
    pf = pfs.get_pf_hash(*args)
    return pf

